# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_mlnx.internal.sys_ops import impl_os
from networking_mlnx import privsep

# for now we have only one implementation
__sys_ops = impl_os.OsCommand()


@privsep.admin_pctx.entrypoint
def sys_write(path, value):
    """Write to sysfs path (/sys/...)

    :param path: sysfs path
    :param value: value to write
    :raises InvalidSystemPathError: provided path is not under sysfs
    :raises SystemOperationError: an error occured when opening/writing
                                  to file.
    """
    __sys_ops.sys_write(path, value)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_mlnx._i18n import _


class NetworkInterfaceNotFound(RuntimeError):
    message = _("Network interface %(device)s not found")

    def __init__(self, message=None, device=None):
        message = message or self.message % {'device': device}
        super(NetworkInterfaceNotFound, self).__init__(message)


class NetlinkRuntimeError(RuntimeError):
    message = _("Failed to execute Netlink command. %(e)s")

    def __init__(self, e):
        message = self.message % {'e': str(e)}
        super(NetlinkRuntimeError, self).__init__(message)


class NetlinkAttrNotFoundError(RuntimeError):
    message = _("Required netlink attribute was not found. %(e)s")

    def __init__(self, e):
        message = self.message % {'e': str(e)}
        super(NetlinkAttrNotFoundError, self).__init__(message)


class NetlinkUnexpectedAttrValue(RuntimeError):
    message = _("Unexpected Netlink attribute value. %(e)s")

    def __init__(self, e):
        message = self.message % {'e': str(e)}
        super(NetlinkUnexpectedAttrValue, self).__init__(message)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd import eswitch_manager
from networking_mlnx.eswitchd.utils import pci_utils
from networking_mlnx.internal.netdev_ops import constants as netdev_consts


LOG = logging.getLogger(__name__)


class eSwitchHandler(object):

    def __init__(self, fabrics=None):
        """Constructor

        :param fabrics: list of <physnet, pf netdev> tuples
        """
        self.eswitch_mgr = eswitch_manager.ESwitchManager()
        self.pci_utils = pci_utils.pciUtils()
        self.devices = set()
        if fabrics:
            self.add_fabrics(fabrics)

    def add_fabrics(self, fabrics):
        """Add fabrics to ESwitchManager

        :param fabrics: list of <physnet, pf netdev> tuples
        """
        self.eswitch_mgr.discover_devices(fabrics)

    def get_vnics(self, fabrics):
        """Get assigned vnics

        :param fabrics: list of physnets to get attached vnics from or None
                        to get vnics from all physnets.
        :return: dict, keyed by the PCI slot of the VNIC
        """
        vnics = {}
        dev_infos = self.eswitch_mgr.get_assigned_devices_info(fabrics)
        for dev_info in dev_infos:
            mac, slot = dev_info
            vnics[slot] = {"mac": mac, "pci_slot": slot}
        LOG.info("vnics are %s", vnics)
        return vnics

    def plug_nic(self, fabric, device_id, vnic_mac, pci_slot):
        # TODO(adrianc): fabric and device_id are no longer used, we should
        # clean them up.
        self.eswitch_mgr.set_device_ib_mac(pci_slot, vnic_mac)
        return pci_slot

    def delete_port(self, fabric, vnic_mac):
        pci_slot = self.eswitch_mgr.get_device_from_mac(vnic_mac, fabric)
        if pci_slot is None:
            LOG.warning("MAC address %s not found in eSwitch for Fabric: %s",
                        vnic_mac, fabric)
        else:
            self.eswitch_mgr.set_device_ib_mac(pci_slot)
        return pci_slot

    def port_release(self, fabric, pci_slot):
        """Cleanup port from any additional configurations
        (excluding mac, admin state)

        :param fabric: physnet
        :param pci_slot: VF PCI address
        :return: None - no cleanup performed.
                 True - cleanup was successful.
                 False - cleanup failed.
        """
        # cleanup Pkey(vlan) (Relevant to ConnectX3 only)
        ret = self.set_vlan(
            fabric, pci_slot, constants.UNTAGGED_VLAN_ID)
        # NOTE(adrianc) port_down() does basically nothing.
        # TODO(adrianc): revisite if we want to remove it.
        # self.port_down(fabric, pci_slot)
        return ret

    def port_up(self, fabric, pci_slot):
        # TODO(adrianc): remove un-needed fabric arg
        self.eswitch_mgr.set_device_admin_state(
            pci_slot, netdev_consts.ADMIN_STATE_UP)

    def port_down(self, fabric, pci_slot):
        # TODO(adrianc): remove un-needed fabric arg
        self.eswitch_mgr.set_device_admin_state(
            pci_slot, netdev_consts.ADMIN_STATE_DOWN)

    def set_vlan(self, fabric, pci_slot, vlan):
        if pci_slot:
            try:
                self.eswitch_mgr.config_vlan_ib(pci_slot, vlan)
                return True
            except RuntimeError as e:
                LOG.error('Set VLAN operation failed. %s', str(e))
        else:
            LOG.debug('No PCI device provided.')
        return False

    def get_eswitch_tables(self, fabrics):
        tables = {}
        # TODO(adrianc): This functionality will be removed in the future,
        # for now just retrieve a table with assigned devices info.
        for fabric in fabrics:
            # Build table for fabric.
            dev_infos = self.eswitch_mgr.get_assigned_devices_info([fabric])
            table = [['MAC', 'PCI']]
            for dev_info in dev_infos:
                mac, slot = dev_info
                table.append([mac, slot])
            tables[fabric] = table
        return tables



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glob
import re

from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.common import exceptions
from networking_mlnx.eswitchd.utils import pci_utils
from networking_mlnx.internal.netdev_ops import api as net_dev_api
from networking_mlnx.internal.netdev_ops import constants as netdev_const
from networking_mlnx.internal.netdev_ops import exceptions as api_exceptions
from networking_mlnx.internal.sys_ops import api as sys_api

LOG = logging.getLogger(__name__)


class IbUtils(object):
    INVALID_PKEY = 'none'
    DEFAULT_PKEY_IDX = 0
    PARTIAL_PKEY_IDX = 1
    DEFAULT_MASK = 0x7fff
    DEFAULT_PKEY = 0xffff
    PKEYS_PATH = "/sys/class/infiniband/%s/ports/%s/pkeys/*"
    GUID_FMT_MLNX4 = r"^[0-9a-fA-F]{16}$"
    GUID_FMT_MLNX5 = r"^([0-9a-fA-F]{2}:){7}([0-9A-Fa-f]{2})$"

    def _config_vf_pkey(self, ppkey_idx, pkey_idx,
                        pf_mlx_dev, vf_pci_id, hca_port):
        path = constants.MLNX4_PKEY_INDEX_PATH % (pf_mlx_dev, vf_pci_id,
                                                  hca_port, pkey_idx)
        sys_api.sys_write(path, ppkey_idx)

    def _get_guid_idx_mlx4(self, pf_mlx_dev, dev, hca_port):
        path = constants.MLNX4_GUID_INDEX_PATH % (pf_mlx_dev, dev, hca_port)
        with open(path) as fd:
            idx = fd.readline().strip()
        return int(idx)

    def _get_guid_from_mac(self, mac, device_type):
        guid = None
        if device_type == constants.MLNX4_DEVICE_TYPE:
            if mac is None:
                guid = constants.MLNX4_INVALID_GUID
            else:
                mac = mac.replace(':', '')
                prefix = mac[:6]
                suffix = mac[6:]
                guid = prefix + '0000' + suffix
        elif (device_type == constants.MLNX5_DEVICE_TYPE):
            if mac is None:
                guid = constants.MLNX5_INVALID_GUID
            else:
                prefix = mac[:9]
                suffix = mac[9:]
                guid = prefix + '00:00:' + suffix
        return guid

    def _get_mac_from_guid(self, guid):
        """Truncate GUID to mac by removing the 4th and 5th bytes.

        :param guid: str formatted in either: xxxxxxxxxxxxxxxx or
                     xx:xx:xx:xx:xx:xx:xx:xx where 'x' is a hexadecimal digit
        :return: mac
        """
        if re.match(IbUtils.GUID_FMT_MLNX4, guid):
            mac = ":".join(re.findall('..?', guid[:6] + guid[-6:]))
        elif re.match(IbUtils.GUID_FMT_MLNX5, guid):
            mac = guid[:8] + guid[-9:]
        else:
            raise exceptions.InvalidGUIDFormatException(guid)
        return mac

    def get_vfs_macs_ib(self, pf_net_name, pf_mlx_name, hca_port, vf_idxs,
                        type):
        """Get assigned Infiniband mac address for VFs

        :param pf_net_name: PF net device name
        :param pf_mlx_name: PF IB device name
        :param hca_port: hca port number
        :param vf_idxs: list of VF indexes to get mac address
        :param type: PF device type, one of [constants.MLNX4_DEVICE_TYPE,
                    constants.MLNX5_DEVICE_TYPE]
        :return: mapping between VF index and mac
        """
        macs_map = {}
        if type == constants.MLNX4_DEVICE_TYPE:
            macs_map.update(
                self._get_vfs_macs_ib_mlnx4(pf_net_name, pf_mlx_name,
                                            hca_port, vf_idxs))
        elif type == constants.MLNX5_DEVICE_TYPE:
            macs_map.update(
                self._get_vfs_macs_ib_mlnx5(pf_net_name, pf_mlx_name,
                                            vf_idxs))
        return macs_map

    def _get_gid_to_vf_idx_mapping(self, pf_ib_dev, hca_port, vf_idxs):
        """Get mapping between VF GID index and VF PCI index

        :param pf_ib_dev: PF IB device name
        :param hca_port: hca port number
        :param vf_idxs: list of vf indexes
        :return: dict mapping between gid index and VF PCI index
        """
        pciu = pci_utils.pciUtils()
        pf_pci_addr = pciu.get_pci_from_ib_dev(pf_ib_dev)
        mapping = {}
        for vf_idx in vf_idxs:
            vf_pci_addr = pciu.get_vf_from_vf_idx(pf_pci_addr, vf_idx)
            vf_gid = self._get_guid_idx_mlx4(pf_ib_dev, vf_pci_addr, hca_port)
            mapping[vf_gid] = vf_idx
        return mapping

    def _get_vfs_ib_mac_netdev_api(self, pf_net_name, vf_idxs):
        try:
            macs_map = {}
            for vf_idx in vf_idxs:
                guid = net_dev_api.get_vf_guid(pf_net_name, vf_idx)
                macs_map[vf_idx] = self._get_mac_from_guid(guid)
            return macs_map
        except api_exceptions.NetlinkAttrNotFoundError:
            return None

    def _get_vfs_macs_ib_mlnx4(self, pf_net_name, pf_mlx_name, hca_port,
                               vf_idxs):
        # NOTE(adrianc): In contrast to MLNX5 devices, it is not expected
        # that MLNX4 devices will support GET vf GUIDs via netlink, hence we
        # do not attempt to get them via netdev API (netlink impl).
        macs_map = {}
        gid_idx_to_vf_idx = self._get_gid_to_vf_idx_mapping(
            pf_mlx_name, hca_port, vf_idxs)
        gid_idxs = gid_idx_to_vf_idx.keys()
        guids_path = constants.MLNX4_ADMIN_GUID_PATH % (pf_mlx_name,
                                                        hca_port,
                                                        '[1-9]*')
        paths = glob.glob(guids_path)
        for path in paths:
            gid_index = int(path.split('/')[-1])
            if gid_index not in gid_idxs:
                continue
            with open(path) as f:
                guid = f.readline().strip()
                if guid == constants.MLNX4_INVALID_GUID:
                    mac = constants.INVALID_MAC
                else:
                    mac = self._get_mac_from_guid(guid)
                macs_map[gid_idx_to_vf_idx[gid_index]] = mac
        return macs_map

    def _get_vfs_macs_ib_mlnx5(self, pf_net_name, pf_mlx_name, vf_idxs):
        macs_map = self._get_vfs_ib_mac_netdev_api(pf_net_name, vf_idxs)
        # TODO(adrianc): The logic below should be removed once major distros
        # have kernel based on 5.5.0 or newer.
        if macs_map is None:
            LOG.debug("Failed to get vf guid via netdev API, "
                     "attempting to get vf guid via sysfs.")
            macs_map = {}
            for vf_idx in vf_idxs:
                guid_path = (
                    constants.MLNX5_GUID_NODE_PATH % {'module': pf_mlx_name,
                                                      'vf_num': int(vf_idx)})
                with open(guid_path) as f:
                    guid = f.readline().strip()
                    mac = self._get_mac_from_guid(guid)
                macs_map[vf_idx] = mac
        return macs_map

    def config_vf_mac_address(self, pf_net_dev, pf_mlx_dev, vf_idx,
                              vf_pci_slot, vf_device_type, hca_port=None,
                              mac=None):
        """Configure VF infiniband MAC address

        :param pf_net_dev: PF net device name
        :param pf_mlx_dev: PF mlx device name
        :param vf_idx: VF index
        :param vf_pci_slot: VF PCI slot
        :param vf_device_type: VF device type, one of
                               (constants.MLNX4_DEVICE_TYPE,
                               constants.MLNX5_DEVICE_TYPE)
        :param hca_port: HCA port number (ConnectX3 family only)
        :param mac: mac address
        """
        vguid = self._get_guid_from_mac(mac, vf_device_type)
        if vf_device_type == constants.MLNX4_DEVICE_TYPE:
            self._config_vf_mac_address_mlnx4(pf_mlx_dev, vf_pci_slot,
                                              hca_port, vguid)
        elif (vf_device_type == constants.MLNX5_DEVICE_TYPE):
            self._config_vf_mac_address_mlnx5(pf_net_dev, pf_mlx_dev, vf_idx,
                                              vf_pci_slot, vguid)
        else:
            LOG.error("Unsupported vf device type: %s ", vf_device_type)

    def _config_vf_mac_address_mlnx4(self, pf_mlx_dev, vf_pci_slot, hca_port,
                                     vguid):
        # NOTE(adrianc): In contrast to MLNX5 devices, it is not expected
        # that MLNX4 devices will support SET vf GUID via netlink, hence we
        # do not attempt to set them via netdev API (netlink impl).

        self._config_vf_pkey(
            IbUtils.INVALID_PKEY, IbUtils.DEFAULT_PKEY_IDX, pf_mlx_dev,
            vf_pci_slot, hca_port)

        guid_idx = self._get_guid_idx_mlx4(pf_mlx_dev, vf_pci_slot,
                                           hca_port)
        path = constants.MLNX4_ADMIN_GUID_PATH % (pf_mlx_dev, hca_port,
                                                  guid_idx)
        sys_api.sys_write(path, vguid)

        ppkey_idx = self._get_pkey_idx(
            IbUtils.DEFAULT_PKEY, pf_mlx_dev, hca_port)
        if ppkey_idx is not None:
            self._config_vf_pkey(
                ppkey_idx, IbUtils.PARTIAL_PKEY_IDX, pf_mlx_dev, vf_pci_slot,
                hca_port)
        else:
            LOG.error("Can't find partial management pkey for "
                      "%(pf)s:%(dev)s", {'pf': pf_mlx_dev, 'dev': vf_pci_slot})

    def _set_vf_guid_sysfs_mlnx5(self, guid, pf_mlx_dev, vf_idx):
        guid_node = constants.MLNX5_GUID_NODE_PATH % {'module': pf_mlx_dev,
                                                      'vf_num': int(vf_idx)}
        guid_port = constants.MLNX5_GUID_PORT_PATH % {'module': pf_mlx_dev,
                                                      'vf_num': int(vf_idx)}
        for path in (guid_node, guid_port):
            sys_api.sys_write(path, guid)

    def _config_vf_mac_address_mlnx5(self, pf_net_dev, pf_mlx_dev, vf_idx,
                                     vf_pci_slot, vguid):
        try:
            net_dev_api.set_vf_guid(pf_net_dev, vf_idx, vguid)
        except api_exceptions.NetlinkRuntimeError as e:
            LOG.debug("Failed to set vf guid via netdev API. "
                     "%s, attempting to set vf guid via sysfs", str(e))
            self._set_vf_guid_sysfs_mlnx5(vguid, pf_mlx_dev, vf_idx)

        if vguid == constants.MLNX5_INVALID_GUID:
            net_dev_api.set_vf_admin_state(
                pf_net_dev, vf_idx, netdev_const.ADMIN_STATE_DOWN)
            sys_api.sys_write(constants.UNBIND_PATH, vf_pci_slot)
            sys_api.sys_write(constants.BIND_PATH, vf_pci_slot)

    def config_vlan_ib(self, pf_mlx_dev, vf_pci_slot, vf_device_type, hca_port,
                       vlan):
        if vf_device_type == constants.MLNX4_DEVICE_TYPE:
            self._config_vlan_ib_mlnx4(pf_mlx_dev, vf_pci_slot, hca_port, vlan)
        elif vf_device_type == constants.MLNX5_DEVICE_TYPE:
            pass
        else:
            LOG.error("Unsupported vf device type: %s ", vf_device_type)

    def _config_vlan_ib_mlnx4(self, pf_mlx_dev, vf_pci_slot, hca_port, vlan):
        if vlan == 0:
            ppkey_idx = self._get_pkey_idx(
                IbUtils.DEFAULT_PKEY, pf_mlx_dev, hca_port)
            if ppkey_idx is not None:
                self._config_vf_pkey(
                    ppkey_idx, IbUtils.DEFAULT_PKEY_IDX, pf_mlx_dev,
                    vf_pci_slot, hca_port)
        else:
            ppkey_idx = self._get_pkey_idx(vlan, pf_mlx_dev, hca_port)
            if ppkey_idx is not None:
                self._config_vf_pkey(
                    ppkey_idx, IbUtils.DEFAULT_PKEY_IDX, pf_mlx_dev,
                    vf_pci_slot, hca_port)

    def _get_pkey_idx(self, vlan, pf_mlx_dev, hca_port):
        """get PKEY index for provided vlan ID

        :param vlan: Vlan ID
        :param pf_mlx_dev: PF mlx device
        :param hca_port: HCA port number
        :return: matching pkey index or None
        """
        paths = IbUtils.PKEYS_PATH % (pf_mlx_dev, hca_port)
        for path in glob.glob(paths):
            fd = open(path)
            pkey = fd.readline()
            fd.close()
            # the MSB in pkey is the membership bit ( 0 - partial, 1 - full)
            # the other 15 bit are the number of the pkey
            # so we want to remove the 16th bit when compare pkey file
            # to the vlan (pkey) we are looking for
            is_match = (int(pkey, 16) & IbUtils.DEFAULT_MASK == vlan &
                        IbUtils.DEFAULT_MASK)
            if is_match:
                return int(path.split('/')[-1])
        return None



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.agent.linux import interface as n_interface
from neutron.agent.linux import ip_lib as n_ip_lib
from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_service import loopingcall
from oslo_utils.fixture import uuidsentinel as uuids
import six

from networking_mlnx.linux.interface_drivers import constants
from networking_mlnx.linux.interface_drivers import interface
from networking_mlnx.linux.interface_drivers import network_cache
from networking_mlnx.linux import ip_lib
from networking_mlnx.tests import base

network_db = {
    uuids.vlan_net: {'id': uuids.vlan_net,
                     constants.SEGMENTATION_ID: "14",
                     constants.PHYSICAL_NETWORK: "datacenter"},
    uuids.flat_net: {'id': uuids.flat_net,
                     constants.SEGMENTATION_ID: None,
                     constants.PHYSICAL_NETWORK: "ib-physnet"},
    uuids.vxlan_net: {'id': uuids.vxlan_net,
                      constants.SEGMENTATION_ID: 900,
                      constants.PHYSICAL_NETWORK: None}
}


class TestIPoIBInterfaceDriver(base.TestCase):
    def _get_networks_cb(self, filters=None, fields=None):
        # Assert calls to callback with single filter and single field
        self.assertEqual(['id'], list(filters.keys()))
        self.assertEqual(1, len(filters["id"]))
        self.assertEqual([constants.SEGMENTATION_ID], fields)
        net_id = filters["id"][0]
        return [network_db[net_id]]

    def setUp(self):
        super(TestIPoIBInterfaceDriver, self).setUp()
        self.root_dev = "ib0"
        cfg.CONF.set_override('ipoib_physical_interface', self.root_dev)
        with mock.patch.object(
                n_interface.LinuxInterfaceDriver, "__init__") as init_mock, \
                mock.patch.object(n_ip_lib, "device_exists"):
            # TODO(adrianc): Bump test-requirements neutron version to Train
            # and remove the __init__ mock above.
            def custom_init(*args, **kwargs):
                pass
            init_mock.side_effect = custom_init
            self.driver = interface.IPoIBInterfaceDriver(
                cfg.CONF, get_networks_callback=self._get_networks_cb)

    def _mock_ipoib_wrapper(self):
        patcher = mock.patch.object(ip_lib, "IPoIBWrapper")
        ipoib_cls_mock = patcher.start()
        ip_mock_inst = mock.MagicMock()
        ipoib_cls_mock.return_value = ip_mock_inst
        self.addCleanup(patcher.stop)
        return ip_mock_inst

    def _test_plug_new(self, net_id, dev_name, link_up=True):
        ip_mock = self._mock_ipoib_wrapper()
        ip_dev_mock = mock.Mock()
        ip_mock.add_ipoib.return_value = ip_dev_mock
        self.driver.plug_new(net_id, uuids.port_id, dev_name,
                            None, link_up)
        if net_id == uuids.flat_net:
            ip_mock.add_ipoib.assert_called_with(
                dev_name, self.root_dev, None)
        else:
            ip_mock.add_ipoib.assert_called_with(
                dev_name, self.root_dev,
                network_db[uuids.vlan_net][constants.SEGMENTATION_ID])
        if link_up:
            ip_dev_mock.link.set_up.asset_called_once()
        else:
            ip_dev_mock.link.asset_not_called()

    def test_plug_new_vlan_network(self):
        self._test_plug_new(uuids.vlan_net, "my-ipoib-netdev")

    def test_plug_new_flat_network(self):
        self._test_plug_new(uuids.flat_net, "my-ipoib-netdev")

    def test_plug_new_link_unchanged(self):
        self._test_plug_new(uuids.vlan_net, "my-ipoib-netdev", False)

    @mock.patch("networking_mlnx.linux.interface_drivers.interface.LOG")
    def test_plug_new_ip_lib_raises(self, log_mock):
        ip_mock = self._mock_ipoib_wrapper()
        ip_mock.add_ipoib.side_effect = RuntimeError("Unexpected!")
        self.driver.plug_new(
            uuids.vlan_net, uuids.port_a, "my-ipoib-netdev", None)
        log_mock.error.assert_called_once()

    def test_unplug(self):
        ip_mock = self._mock_ipoib_wrapper()
        self.driver.unplug("my-ipoib-netdev", None, namespace="test-ns")
        ip_mock.del_ipoib.assert_called_with("my-ipoib-netdev")

    @mock.patch("networking_mlnx.linux.interface_drivers.interface.LOG")
    def test_unplug_ip_lib_raises(self, log_mock):
        ip_mock = self._mock_ipoib_wrapper()
        ip_mock.del_ipoib.side_effect = RuntimeError("Unexpected!")
        self.driver.unplug("my-ipoib-netdev")
        log_mock.error.assert_called_once()


class TestMultiInterfaceDriver(base.TestCase):
    def _get_networks_cb(self, filters=None, fields=None):
        if 'id' in filters.keys():
            net_id = filters["id"][0]
            return [network_db[net_id]]
        return [list(network_db.values())]

    def setUp(self):
        super(TestMultiInterfaceDriver, self).setUp()
        interface.MultiInterfaceDriver.network_cache = None
        interface.MultiInterfaceDriver._cache_init_lock = mock.MagicMock()
        self.fields = ['id', constants.PHYSICAL_NETWORK,
                       constants.SEGMENTATION_ID]
        # TODO(adrianc): Bump test-requirements neutron version to Train
        # and remove the __init__ mock above.
        self.init_patcher = mock.patch.object(
            n_interface.LinuxInterfaceDriver, "__init__")
        self.init_patcher.start()

    def tearDown(self):
        super(TestMultiInterfaceDriver, self).tearDown()
        self.init_patcher.stop()

    def _get_mock_drivers(self):
        drivers = {'datacenter': mock.Mock(),
                   'ib-physnet': mock.Mock(),
                   'nil': mock.Mock()}
        drivers['datacenter']._interface_kind_ = 'openvswitch'
        drivers['ib-physnet']._interface_kind_ = 'ipoib'
        drivers['nil']._interface_kind_ = 'openvswitch'
        return drivers

    def _init_network_cache(self):
        get_networks_mock = mock.Mock()
        get_networks_mock.return_value = {"IfYouSeeThis": "ThenYouBrokeIt"}
        fields = ['id', constants.PHYSICAL_NETWORK, constants.SEGMENTATION_ID]
        interface.MultiInterfaceDriver.network_cache = (
            network_cache.NetworkCache(get_networks_mock, fields))
        for net in six.itervalues(network_db):
            interface.MultiInterfaceDriver.network_cache.put(net['id'], net)

    @mock.patch.object(loopingcall, "FixedIntervalLoopingCall")
    @mock.patch.object(network_cache, "SafeNetworkCache")
    def test__init_network_cache(self, net_cache_cls_mock, looping_mock):
        conf = mock.Mock()
        conf.enable_multi_interface_driver_cache_maintenance = False
        interface.MultiInterfaceDriver._init_network_cache(
            conf, self._get_networks_cb, self.fields)
        net_cache_cls_mock.assert_called_once_with(self._get_networks_cb,
                                                   self.fields)
        self.assertFalse(looping_mock.called)
        # Make sure consecutive calls dont re-initialize network cache
        net_cache_cls_mock.reset_mock()
        interface.MultiInterfaceDriver._init_network_cache(
            conf, self._get_networks_cb, self.fields)
        self.assertFalse(net_cache_cls_mock.called)

    @mock.patch.object(loopingcall, "FixedIntervalLoopingCall")
    @mock.patch.object(network_cache, "SafeNetworkCache")
    def test__init_network_cache_with_cache_maintenance(self,
                                                        net_cache_cls_mock,
                                                        looping_mock):
        conf = mock.Mock()
        conf.enable_multi_interface_driver_cache_maintenance = True
        net_cache_mock = mock.Mock()
        net_cache_cls_mock.return_value = net_cache_mock
        loop_obj = mock.Mock()
        looping_mock.return_value = loop_obj
        interface.MultiInterfaceDriver._init_network_cache(
            conf, self._get_networks_cb, self.fields)
        looping_mock.assert_called_once_with(
            net_cache_mock.remove_stale_networks)
        loop_obj.start.called_once_with()
        # Make sure consecutive calls dont re-spawn cleanup thread
        looping_mock.reset_mock()
        loop_obj.start.reset_mock()
        looping_mock.return_value = loop_obj
        interface.MultiInterfaceDriver._init_network_cache(
            conf, self._get_networks_cb, self.fields)
        looping_mock.assert_not_called()
        loop_obj.start.assert_not_called()

    @mock.patch.object(interface.MultiInterfaceDriver, "_process_driver_obj")
    @mock.patch.object(runtime, 'load_class_by_alias_or_classname')
    def test_load_interface_driver_mappings(self, load_cls_mock, process_mock):
        loaded_intf_drivers = []

        def load_cls(namespace, name):
            loaded_intf_drivers.append(name)
            return mock.Mock()

        load_cls_mock.side_effect = load_cls
        conf = mock.Mock()
        conf.multi_interface_driver_mappings = (
            "physnetA:openvswitch,physnetB:ipoib")
        mapping = (
            interface.MultiInterfaceDriver.load_interface_driver_mappings(
                conf, get_networks_callback=mock.Mock()))
        self.assertEqual(set(["physnetA", "physnetB"]), set(mapping.keys()))
        self.assertEqual(set(["openvswitch", "ipoib"]),
                         set(loaded_intf_drivers))

    def _check_drivers(self, first_driver_spec, second_driver_spec,
                       should_raise=False):
        type, kind = first_driver_spec
        first_mock = mock.Mock(spec=type)
        first_mock._interface_kind_ = kind
        type, kind = second_driver_spec
        second_mock = mock.Mock(spec=type)
        second_mock._interface_kind_ = kind
        mapping = {"physnetA": first_mock, "physnetB": second_mock}
        if should_raise:
            self.assertRaises(SystemExit,
                              interface.MultiInterfaceDriver._check_drivers,
                              mapping)
        else:
            interface.MultiInterfaceDriver._check_drivers(mapping)

    def test__check_drivers_different_type(self):
        self._check_drivers((n_interface.OVSInterfaceDriver, 'openvswitch'),
                       (interface.IPoIBInterfaceDriver, 'ipoib'),
                       False)

    def test__check_drivers_same_type(self):
        self._check_drivers((n_interface.OVSInterfaceDriver, 'openvswitch'),
                       (n_interface.OVSInterfaceDriver, 'openvswitch'),
                       False)

    def test__check_drivers_raises(self):
        self._check_drivers((n_interface.BridgeInterfaceDriver, 'veth'),
                            (n_interface.OVSInterfaceDriver, 'veth'),
                            True)

    def _check_process_driver_obj_ovs(self, conf, expected_kind):
        ovs_driver = mock.Mock(spec=n_interface.OVSInterfaceDriver)
        ovs_driver.conf = conf
        ovs_driver = interface.MultiInterfaceDriver._process_driver_obj(
            ovs_driver)
        self.assertTrue(hasattr(ovs_driver, "_interface_kind_"))
        self.assertEqual(expected_kind, ovs_driver._interface_kind_)

    def test__process_driver_obj_ovs(self):
        conf = mock.Mock()
        conf.ovs_use_veth = False
        self._check_process_driver_obj_ovs(conf, "openvswitch")

    def test__process_driver_obj_ovs_veth(self):
        conf = mock.Mock()
        conf.ovs_use_veth = True
        self._check_process_driver_obj_ovs(conf, "veth")

    @mock.patch.object(n_ip_lib, "device_exists")
    def test__process_ipoib_driver_obj(self, dev_exist_mock):
        conf = mock.Mock()
        conf.ipoib_physical_interface = "ib0"
        ipoib_driver = interface.IPoIBInterfaceDriver(
            conf, get_networks_callback=mock.Mock())
        ipoib_driver = interface.MultiInterfaceDriver._process_driver_obj(
            ipoib_driver)
        self.assertTrue(hasattr(ipoib_driver, "_interface_kind_"))
        self.assertEqual('ipoib', ipoib_driver._interface_kind_)

    def test__process_unknown_driver_obj(self):
        unknown_driver = mock.Mock(spec=['conf'])
        unknown_driver = interface.MultiInterfaceDriver._process_driver_obj(
            unknown_driver)
        self.assertTrue(hasattr(unknown_driver, "_interface_kind_"))
        self.assertEqual('unknown', unknown_driver._interface_kind_)

    def test__get_networks_from_cache_all_nets(self):
        self._init_network_cache()
        nets = interface.MultiInterfaceDriver._get_networks_from_cache(
            filters=None, fields=None)
        for net in nets:
            # assert net object are identical
            self.assertDictEqual(network_db[net['id']], net)
        # assert all objects where retrieved
        self.assertEqual(set(network_db.keys()), set(
            [net['id'] for net in nets]))

    def test__get_networks_from_cache_some_fields(self):
        self._init_network_cache()
        some_fields = ['id', constants.PHYSICAL_NETWORK]
        nets = interface.MultiInterfaceDriver._get_networks_from_cache(
            filters=None, fields=some_fields)
        for net in nets:
            # assert retrieved net obj contains only the requested fields
            # 'id' field
            self.assertIn('id', net)
            self.assertIn(net['id'], network_db)
            # 'provider:physical_network' field
            self.assertIn(constants.PHYSICAL_NETWORK, net)
            self.assertEqual(
                network_db[net['id']][constants.PHYSICAL_NETWORK],
                net[constants.PHYSICAL_NETWORK])
            # No additional fields
            self.assertEqual(set(some_fields), set(net.keys()))
        # assert all networks were retrieved
        self.assertEqual(set(network_db.keys()), set(
            [net['id'] for net in nets]))

    def test__get_networks_from_cache_some_filters(self):
        self._init_network_cache()
        nets = interface.MultiInterfaceDriver._get_networks_from_cache(
            filters={'id': [uuids.vlan_net]}, fields=None)
        self.assertEqual(1, len(nets))
        self.assertDictEqual(network_db[uuids.vlan_net], nets[0])

    def test__get_networks_from_cache_invalid_field(self):
        self._init_network_cache()
        nets = interface.MultiInterfaceDriver._get_networks_from_cache(
            fields=['invalid'])
        self.assertEqual([], nets)

    def test__get_networks_from_cache_invalid_filter(self):
        self._init_network_cache()
        nets = interface.MultiInterfaceDriver._get_networks_from_cache(
            filters={'invalid': ['i', 'n', 'v', 'a', 'l', 'i', 'd']})
        self.assertEqual([], nets)

    @mock.patch.object(n_ip_lib, 'IPDevice')
    @mock.patch.object(interface.MultiInterfaceDriver, '_init_network_cache')
    @mock.patch.object(interface.MultiInterfaceDriver,
                       'load_interface_driver_mappings')
    def test__get_driver_for_existing_interface(self, load_mapping_mock,
                                                init_net_cache_mock,
                                                ip_device_mock):
        drivers = self._get_mock_drivers()
        load_mapping_mock.return_value = drivers
        ipd_mock = mock.Mock()
        ipd_mock.link.link_kind = 'ipoib'
        ip_device_mock.return_value = ipd_mock
        multi_driver = interface.MultiInterfaceDriver(mock.Mock, None)
        driver = multi_driver._get_driver_for_existing_interface("dev-name")
        self.assertEqual(drivers['ib-physnet'], driver)
        # in case a matching driver is not found - ensure None is returned
        ipd_mock.link.link_kind = 'veth'
        driver = multi_driver._get_driver_for_existing_interface("dummy-dev")
        self.assertIsNone(driver)

    @mock.patch.object(n_ip_lib, 'device_exists')
    @mock.patch.object(interface.MultiInterfaceDriver,
                       '_get_driver_for_existing_interface')
    @mock.patch.object(interface.MultiInterfaceDriver, '_init_network_cache')
    @mock.patch.object(interface.MultiInterfaceDriver,
                       'load_interface_driver_mappings')
    def _unplug_device(self, device_exists, load_mapping_mock,
                       init_net_cache_mock, get_driver_mock,
                       device_exists_mock):
        device_exists_mock.return_value = device_exists
        mock_driver = mock.Mock()
        get_driver_mock.return_value = mock_driver
        driver = interface.MultiInterfaceDriver(mock.Mock, None)
        driver.unplug("dummy-dev", namespace="dummy-ns")
        if device_exists:
            mock_driver.unplug.assert_called_once_with("dummy-dev", None,
                                                       "dummy-ns", None)
        else:
            mock_driver.unplug.assert_not_called()

    def test_unplug_device_exist(self):
        self._unplug_device(True)

    def test_unplug_device_does_not_exist(self):
        self._unplug_device(False)

    @mock.patch.object(interface.MultiInterfaceDriver,
                       '_get_driver_for_existing_interface')
    @mock.patch.object(interface.MultiInterfaceDriver, '_init_network_cache')
    @mock.patch.object(interface.MultiInterfaceDriver,
                       'load_interface_driver_mappings')
    def test_set_mtu(self, load_mapping_mock, init_net_cache_mock,
                     get_driver_mock):
        driver_mock = mock.Mock()
        get_driver_mock.return_value = driver_mock
        driver = interface.MultiInterfaceDriver(mock.Mock, None)
        driver.set_mtu("dummy-dev", 1500, namespace="dummy-ns", prefix=None)
        driver_mock.set_mtu.assert_called_once_with(
            "dummy-dev", 1500, "dummy-ns", None)

    @mock.patch.object(interface.MultiInterfaceDriver, '_init_network_cache')
    @mock.patch.object(interface.MultiInterfaceDriver,
                       'load_interface_driver_mappings')
    def test_plug_new(self, load_mapping_mock, init_net_cache_mock):
        self._init_network_cache()
        driver = interface.MultiInterfaceDriver(mock.Mock, None)
        driver.drivers = self._get_mock_drivers()
        device_name = 'test-dev'
        mac = 'fa:18:64:2d:07:d5'
        ns = 'test-ns'
        # network with physnet
        driver.plug_new(uuids.vlan_net, uuids.vlan_port, device_name, mac,
                        bridge=None, namespace=ns, prefix=None, mtu=None,
                        link_up=True)
        driver.drivers['datacenter'].plug_new.assert_called_once_with(
            uuids.vlan_net, uuids.vlan_port, device_name, mac, None, ns, None,
            None, True)
        # network without physnet
        driver.plug_new(uuids.vxlan_net, uuids.vxlan_port, device_name, mac,
                        bridge=None, namespace=ns, prefix=None, mtu=None,
                        link_up=True)
        driver.drivers['nil'].plug_new.assert_called_once_with(
            uuids.vxlan_net, uuids.vxlan_port, device_name, mac, None, ns,
            None, None, True)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.privileged.agent.linux import ip_lib as n_privileged

from networking_mlnx.linux import constants
from networking_mlnx.linux import ip_lib
from networking_mlnx.tests import base


class TestIPoIBWrapper(base.TestCase):
    def setUp(self):
        super(TestIPoIBWrapper, self).setUp()
        self.execute_p = mock.patch.object(ip_lib.IPoIBWrapper, '_execute')
        self.execute = self.execute_p.start()
        self.namespace = "test-ns"
        self.ipoib = ip_lib.IPoIBWrapper(namespace=self.namespace)

    def test_segmentation_id_to_pkey_default(self):
        pkey = self.ipoib._segmentation_id_to_pkey(None)
        self.assertEqual(constants.DEFAULT_PKEY, pkey)
        pkey = self.ipoib._segmentation_id_to_pkey(0)
        self.assertEqual(constants.DEFAULT_PKEY, pkey)

    def test_segmentation_id_to_pkey_int_val(self):
        seg_id = 4
        pkey = self.ipoib._segmentation_id_to_pkey(seg_id)
        self.assertEqual(seg_id, pkey)

    def test_segmentation_id_to_pkey_str_val(self):
        seg_id = "14"
        pkey = self.ipoib._segmentation_id_to_pkey(seg_id)
        self.assertEqual(int(seg_id), pkey)

    def test_del_ipoib(self):
        with mock.patch.object(
                n_privileged, 'delete_interface') as del_ifc_mock:
            self.ipoib.del_ipoib("ipoib0")
            del_ifc_mock.assert_called_with("ipoib0", self.namespace)

    def test_add_ipoib(self):
        with mock.patch.object(
                n_privileged, 'create_interface') as create_ifc_mock, \
            mock.patch.object(
                n_privileged, 'set_link_attribute') as set_link_attr_mock:
            name = "ipoib0"
            src_dev = "ib0"
            pkey = 0x4
            ipdev = self.ipoib.add_ipoib(name, src_dev, pkey)
            # Assert device created in default namespace
            create_ifc_mock.assert_called_with(
                name, None, 'ipoib', physical_interface=src_dev, pkey=pkey)
            # Assert device moved to namespace
            set_link_attr_mock.assert_called_with(
                name, None, net_ns_fd=self.namespace)
            self.assertEqual(name, ipdev.name)
            self.assertEqual(self.namespace, ipdev.namespace)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_mlnx.eswitchd.common import exceptions
from networking_mlnx.eswitchd.utils import ib_utils
from networking_mlnx.internal.netdev_ops import exceptions as api_exceptions
from networking_mlnx.tests import base


class TestIbUtils(base.TestCase):

    def setUp(self):
        super(TestIbUtils, self).setUp()
        self.ib_utils = ib_utils.IbUtils()

    def test__get_mac_from_guid_mlx4(self):
        guid = "1122334455667788"
        expected_mac = "11:22:33:66:77:88"
        mac = self.ib_utils._get_mac_from_guid(guid)
        self.assertEqual(expected_mac, mac)

    def test__get_mac_from_guid_mlx5(self):
        guid = "11:22:33:44:55:66:77:88"
        expected_mac = "11:22:33:66:77:88"
        mac = self.ib_utils._get_mac_from_guid(guid)
        self.assertEqual(expected_mac, mac)

    def test__get_mac_from_guid_invalid_guid(self):
        self.assertRaises(exceptions.InvalidGUIDFormatException,
                          self.ib_utils._get_mac_from_guid, "112233445566")
        self.assertRaises(exceptions.InvalidGUIDFormatException,
                          self.ib_utils._get_mac_from_guid,
                          "11:22:33:44:55:66")
        self.assertRaises(exceptions.InvalidGUIDFormatException,
                          self.ib_utils._get_mac_from_guid, "rubbish")

    @mock.patch.object(ib_utils.net_dev_api, "get_vf_guid")
    def test__get_vfs_ib_mac_netlink(self, get_vf_guid_mock):
        def mock_call(pf_ifname, vf_idx):
            guids = ["11:22:33:44:55:66:77:88", "aa:bb:cc:dd:ee:ff:00:11"]
            return guids[vf_idx]
        get_vf_guid_mock.side_effect = mock_call

        mac_map = self.ib_utils._get_vfs_ib_mac_netdev_api("ib0", [0, 1])
        expected_mac_map = {0: "11:22:33:66:77:88",
                            1: "aa:bb:cc:ff:00:11"}
        self.assertDictEqual(expected_mac_map, mac_map)

    @mock.patch.object(ib_utils.net_dev_api, "get_vf_guid",
                       side_effect=api_exceptions.NetlinkAttrNotFoundError(""))
    def test__get_vfs_ib_mac_netlink_fail(self, get_vf_guid_mock):
        mac_map = self.ib_utils._get_vfs_ib_mac_netdev_api("ib0", [0, 1])
        self.assertIsNone(mac_map)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class SystemOps(object):
    """SystemOps consolidates system operations which revolve (mainly) around
    sysfs that may or may not require elevation of privleges.

    """

    @abc.abstractmethod
    def sys_write(self, path, value):
        """Write to sysfs path (/sys/...)

        :param path: sysfs path
        :param value: value to write
        :raises InvalidSystemPathError: provided path is not under sysfs
        :raises SystemOperationError: an error occured when opening/writing
                                      to file.
        """



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron.agent.linux import ip_lib as n_ip_lib
from neutron.privileged.agent.linux import ip_lib as n_privileged

from networking_mlnx.linux import constants


LOG = logging.getLogger(__name__)


class IPoIBWrapper(n_ip_lib.IPWrapper):

    def __init__(self, *args, **kwargs):
        super(IPoIBWrapper, self).__init__(*args, **kwargs)

    def _segmentation_id_to_pkey(self, segmentation_id):
        # NOTE(adrianc): use the default pkey (7fff) in case:
        # 1. network is flat (segmentation_id is None)
        # 2. segmentation_id is 0
        # PKEY length is 15 bits.
        if segmentation_id is None or segmentation_id == 0:
            return constants.DEFAULT_PKEY
        return int(segmentation_id)

    def add_ipoib(self, name, src_dev, segmentation_id=None):
        LOG.debug("Adding IPoIB device: name:%s, src_dev:%s, "
                  "segmentation_id:%s", name, src_dev, segmentation_id)
        pkey = self._segmentation_id_to_pkey(segmentation_id)
        # NOTE(adrianc): ipoib child interface needs to be created in the
        # same namespace as its root device (i.e the physical interface).
        # Create in default namespace and then move.
        n_privileged.create_interface(name,
                                     None,
                                     "ipoib",
                                     physical_interface=src_dev,
                                     pkey=pkey)
        n_privileged.set_link_attribute(
            name, None, net_ns_fd=self.namespace)
        return n_ip_lib.IPDevice(name, namespace=self.namespace)

    def del_ipoib(self, name):
        LOG.debug("Deleting IPoIB device: name:%s", name)
        n_privileged.delete_interface(name, self.namespace)



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron.tests import base
from oslo_config import cfg
import testtools

from networking_mlnx.plugins.ml2.drivers.mlnx.agent import (
    mlnx_eswitch_neutron_agent)
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import exceptions
from networking_mlnx.plugins.ml2.drivers.mlnx.agent import utils


class TestEswichManager(base.BaseTestCase):

    def setUp(self):
        super(TestEswichManager, self).setUp()

        class MockEswitchUtils(object):
            def __init__(self, endpoint, timeout):
                pass

        mock.patch('networking_mlnx.plugins.mlnx.agent.utils.EswitchManager',
                   new=MockEswitchUtils)

        with mock.patch.object(utils, 'zmq'):
            self.manager = mlnx_eswitch_neutron_agent.EswitchManager(
                {}, None, None)

    def test_get_not_exist_port_id(self):
        with testtools.ExpectedException(exceptions.MlnxException):
            self.manager.get_port_id_by_pci('no-such-pci')


class TestMlnxEswitchRpcCallbacks(base.BaseTestCase):

    def setUp(self):
        super(TestMlnxEswitchRpcCallbacks, self).setUp()
        agent = mock.Mock()
        self.rpc_callbacks = \
            mlnx_eswitch_neutron_agent.MlnxEswitchRpcCallbacks(
                'context',
                agent,
                agent)

    def test_port_update(self):
        port = {
            'id': '806f1a00-77cf-4331-9c2d-93e4c321f31d',
            'mac_address': '10:20:30:40:50:60',
            'binding:profile': {'pci_slot': '0000:05:00.3'}
        }
        add_port_update = self.rpc_callbacks.agent.add_port_update
        self.rpc_callbacks.port_update('context', port=port)
        add_port_update.assert_called_once_with(
            (port['mac_address'], port['binding:profile']['pci_slot']))

    def test_port_update_no_pci(self):
        port = {
            'id': '806f1a00-77cf-4331-9c2d-93e4c321f31d',
            'mac_address': '10:20:30:40:50:60',
            'binding:profile': {}
        }
        add_port_update = self.rpc_callbacks.agent.add_port_update
        self.rpc_callbacks.port_update('context', port=port)
        self.assertFalse(add_port_update.called)


class TestEswitchAgent(base.BaseTestCase):
    VNIC_PORT_1 = ('10:20:30:40:50:60', '0000:06:00.2')
    VNIC_PORT_2 = ('11:21:31:41:51:61', '0000:06:00.3')
    VNIC_PORT_3 = ('12:22:32:42:52:62', '0000:06:00.4')
    VNIC_PORT_4 = ('13:23:33:43:53:63', '0000:06:00.5')
    VNIC_PORT_5 = ('14:24:34:44:54:64', '0000:06:00.6')
    VNIC_PORT_1_INVALID_MAC = ('ff:ff:ff:ff:ff:ff', '0000:06:00.2')
    VNIC_PORT_2_INVALID_MAC = ('00:00:00:00:00:00', '0000:06:00.3')

    def setUp(self):
        super(TestEswitchAgent, self).setUp()
        cfg.CONF.set_default('firewall_driver',
                             'noop',
                             group='SECURITYGROUP')

        class MockFixedIntervalLoopingCall(object):
            def __init__(self, f):
                self.f = f

            def start(self, interval=0):
                self.f()

        mock.patch('neutron.openstack.common.loopingcall.'
                   'FixedIntervalLoopingCall',
                   new=MockFixedIntervalLoopingCall)

        with mock.patch.object(utils, 'zmq'):
            self.agent = mlnx_eswitch_neutron_agent.MlnxEswitchNeutronAgent(
                {})
        self.agent.plugin_rpc = mock.Mock()
        self.agent.context = mock.Mock()
        self.agent.agent_id = mock.Mock()
        self.agent.eswitch = mock.Mock()
        self.agent.eswitch.get_vnics.return_value = []

    def test_treat_devices_added_returns_true_for_missing_device(self):
        attrs = {'get_devices_details_list.side_effect': Exception()}
        self.agent.plugin_rpc.configure_mock(**attrs)
        with mock.patch('networking_mlnx.plugins.ml2.drivers.mlnx.agent.'
                        'mlnx_eswitch_neutron_agent.EswitchManager.'
                        'get_vnics',
                        return_value=[]):
            self.assertTrue(self.agent.treat_devices_added_or_updated([{}]))

    def _mock_treat_devices_added_updated(self, details, func_name):
        """Mock treat devices added.

        :param details: the details to return for the device
        :param func_name: the function that should be called
        :returns: whether the named function was called
        """

        with mock.patch('networking_mlnx.plugins.ml2.drivers.mlnx.agent.'
                        'mlnx_eswitch_neutron_agent.EswitchManager.'
                        'get_vnics',
                        return_value=[]),\
            mock.patch.object(self.agent.plugin_rpc,
                              'get_devices_details_list',
                              return_value=[details]),\
            mock.patch.object(self.agent.plugin_rpc,
                              'update_device_up') as upd_dev_up,\
            mock.patch.object(self.agent.plugin_rpc,
                              'update_device_down') as upd_dev_down,\
            mock.patch.object(self.agent, func_name) as func:
            self.assertFalse(self.agent.treat_devices_added_or_updated(
                [self.VNIC_PORT_1, self.VNIC_PORT_2]))
        return (func.called, upd_dev_up.called, upd_dev_down.called)

    def test_treat_devices_added_updates_known_port(self):
        details = mock.MagicMock()
        details.__contains__.side_effect = lambda x: True
        func, dev_up, dev_down = self._mock_treat_devices_added_updated(
                                 details, 'treat_vif_port')
        self.assertTrue(func)
        self.assertTrue(dev_up)

    def test_treat_devices_added_updates_known_port_admin_down(self):
        details = {'device': '01:02:03:04:05:06',
                   'network_id': '123456789',
                   'network_type': 'vlan',
                   'physical_network': 'default',
                   'segmentation_id': 2,
                   'admin_state_up': False}
        func, dev_up, dev_down = self._mock_treat_devices_added_updated(
                                     details, 'treat_vif_port')
        self.assertFalse(func)
        self.assertFalse(dev_up)
        self.assertTrue(dev_down)

    def test_treat_devices_added_updates_known_port_admin_up(self):
        details = {'port_id': '1234567890',
                   'device': '01:02:03:04:05:06',
                   'profile': {'pci_slot': '0000:06:00.2'},
                   'network_id': '123456789',
                   'network_type': 'vlan',
                   'physical_network': 'default',
                   'segmentation_id': 2,
                   'admin_state_up': False}
        func, dev_up, dev_down = self._mock_treat_devices_added_updated(
                                 details, 'treat_vif_port')
        self.assertTrue(func)
        self.assertTrue(dev_up)
        self.assertFalse(dev_down)

    def test_treat_devices_removed_returns_true_for_missing_device(self):
        with mock.patch.object(self.agent.plugin_rpc, 'update_device_down',
                               side_effect=Exception()):
            self.assertTrue(self.agent.treat_devices_removed(
                set([self.VNIC_PORT_1])))

    def test_treat_devices_removed_releases_port(self):
        details = dict(exists=False)
        with mock.patch.object(self.agent.plugin_rpc, 'update_device_down',
                               return_value=details):
            with mock.patch.object(self.agent.eswitch,
                                   'port_release') as port_release:
                self.assertFalse(self.agent.treat_devices_removed(
                    set([self.VNIC_PORT_1])))
                self.assertTrue(port_release.called)

    def _test_process_network_ports(self, port_info):
        with mock.patch.object(self.agent,
                               'treat_devices_added_or_updated',
                               return_value=False) as device_added_updated,\
            mock.patch.object(self.agent, 'treat_devices_removed',
                              return_value=False) as device_removed:
            self.assertFalse(self.agent.process_network_ports(port_info))
            device_added_updated.assert_called_once_with(
                port_info['added'] | port_info['updated'])
            device_removed.assert_called_once_with(port_info['removed'])

    def test_process_network_ports(self):
        self._test_process_network_ports(
            {'current': set(self.VNIC_PORT_1),
             'updated': set(),
             'added': set(self.VNIC_PORT_2),
             'removed': set(self.VNIC_PORT_3)})

    def test_process_network_ports_with_updated_ports(self):
        self._test_process_network_ports(
            {'current': set(self.VNIC_PORT_1),
             'updated': set(self.VNIC_PORT_2),
             'added': set(self.VNIC_PORT_3),
             'removed': set(self.VNIC_PORT_4)})

    def test_add_port_update(self):
        self.agent.add_port_update(self.VNIC_PORT_1)
        self.assertEqual(set([self.VNIC_PORT_1]), self.agent.updated_ports)

    def _mock_scan_ports(self, vif_port_set, previous,
                         updated_ports, sync=False):
        self.agent.updated_ports = updated_ports
        with mock.patch.object(self.agent.eswitch, 'get_vnics',
                               return_value=vif_port_set):
            return self.agent.scan_ports(previous, sync)

    def test_scan_ports_return_current_for_unchanged_ports(self):
        vif_port_set = set([self.VNIC_PORT_1, self.VNIC_PORT_2])
        previous = dict(
            current=set([self.VNIC_PORT_1, self.VNIC_PORT_2]), added=set(),
            removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set(),
                        removed=set(), updated=set())
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set())
        self.assertEqual(expected, actual)

    def test_scan_ports_return_port_changes(self):
        vif_port_set = set([self.VNIC_PORT_1, self.VNIC_PORT_3])
        previous = dict(current=set([self.VNIC_PORT_1, self.VNIC_PORT_2]),
                        added=set(), removed=set(), updated=set())
        expected = dict(current=vif_port_set, added=set([self.VNIC_PORT_3]),
                        removed=set([self.VNIC_PORT_2]), updated=set())
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set())
        self.assertEqual(expected, actual)

    def test_scan_ports_with_updated_ports(self):
        vif_port_set = set(
            [self.VNIC_PORT_1, self.VNIC_PORT_3, self.VNIC_PORT_4])
        previous = dict(
            current=set([self.VNIC_PORT_1, self.VNIC_PORT_2,
                         self.VNIC_PORT_4]),
            added=set(),
            removed=set(),
            updated=set())
        expected = dict(
            current=vif_port_set,
            added=set([self.VNIC_PORT_3]),
            removed=set([self.VNIC_PORT_2]),
            updated=set([self.VNIC_PORT_4]))
        actual = self._mock_scan_ports(vif_port_set,
                                       previous, set([self.VNIC_PORT_4]))
        self.assertEqual(expected, actual)

    def test_scan_ports_with_unknown_updated_ports(self):
        vif_port_set = set(
            [self.VNIC_PORT_1, self.VNIC_PORT_3, self.VNIC_PORT_4])
        previous = dict(
            current=set([self.VNIC_PORT_1, self.VNIC_PORT_2,
                         self.VNIC_PORT_4]),
            added=set(),
            removed=set(),
            updated=set())
        expected = dict(current=vif_port_set,
                        added=set([self.VNIC_PORT_3]),
                        removed=set([self.VNIC_PORT_2]),
                        updated=set([self.VNIC_PORT_4]))
        actual = self._mock_scan_ports(
            vif_port_set, previous, updated_ports=set(
                [self.VNIC_PORT_4, self.VNIC_PORT_5]))
        self.assertEqual(expected, actual)

    def test_scan_ports_with_invalid_vif(self):
        vif_port_set = set(
            [self.VNIC_PORT_1, self.VNIC_PORT_2_INVALID_MAC, self.VNIC_PORT_4])
        previous = dict(
            current=set(
                [self.VNIC_PORT_1, self.VNIC_PORT_3, self.VNIC_PORT_4]),
            added=set(),
            removed=set(),
            updated=set())
        expected = dict(current=vif_port_set - {self.VNIC_PORT_2_INVALID_MAC},
                        added=set(),
                        removed=set([self.VNIC_PORT_3]),
                        updated=set([self.VNIC_PORT_1]))
        actual = self._mock_scan_ports(
            vif_port_set, previous, updated_ports=set([self.VNIC_PORT_1]))
        self.assertEqual(expected, actual)

    def test_fix_eswitchd_vnic_macs(self):
        vif_port_set = {self.VNIC_PORT_1_INVALID_MAC,
                        self.VNIC_PORT_2_INVALID_MAC, self.VNIC_PORT_3}
        updated_ports_cache = {self.VNIC_PORT_1[1]: self.VNIC_PORT_1[0]}
        fixed_vif_ports = self.agent.fix_eswitchd_vnic_macs(
            vif_port_set, updated_ports_cache)
        expected_vif_port_set = set([self.VNIC_PORT_1, self.VNIC_PORT_3])
        self.assertEqual(expected_vif_port_set, fixed_vif_ports)



#
# Copyright (C) 2016 Intel Corp. Isaku Yamahata <isaku.yamahata@gmail com>
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import mock

from neutron.tests import base

from networking_mlnx.journal import dependency_validations


class DependencyValidationsTestCase(base.DietTestCase):
    _RESOURCE_DUMMY = 'test_type'

    def setUp(self):
        super(DependencyValidationsTestCase, self).setUp()
        mock_validation_map = mock.patch.dict(
            dependency_validations._VALIDATION_MAP)
        mock_validation_map.start()
        self.addCleanup(mock_validation_map.stop)

    def test_register_validator(self):
        mock_session = mock.Mock()
        mock_validator = mock.Mock(return_value=False)
        mock_row = mock.Mock()
        mock_row.object_type = self._RESOURCE_DUMMY
        dependency_validations.register_validator(self._RESOURCE_DUMMY,
                                                  mock_validator)
        valid = dependency_validations.validate(mock_session, mock_row)
        mock_validator.assert_called_once_with(mock_session, mock_row)
        self.assertFalse(valid)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import pyroute2

from networking_mlnx.internal.netdev_ops import exceptions
from networking_mlnx.internal.netdev_ops import netdev_ops_abs

LOG = logging.getLogger(__name__)


class IpCommand(netdev_ops_abs.NetDevOperations):
    """_IpCommand implements NetDevOperations interface using pyroute2

    """
    def set_vf_admin_state(self, pf_ifname, vf_idx, state):
        """Set vf admin state

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :param state: desired admin state as defined in
                      networking_mlnx.internal.netdev_ops.constants
        """
        try:
            ip = pyroute2.IPRoute()
            link_idx = ip.link_lookup(ifname=pf_ifname)[0]
            ip.link(
                'set', index=link_idx, vf={'vf': int(vf_idx),
                                           'link_state': state})
        except IndexError:
            raise exceptions.NetworkInterfaceNotFound(pf_ifname)
        except pyroute2.NetlinkError as e:
            raise exceptions.NetlinkRuntimeError(e)

    def set_link_state(self, ifname, state):
        """Set net device link state

        :param ifname: netdev name
        :param state: desired link state as defined in
                      networking_mlnx.internal.netdev_ops.constants
        """
        try:
            ip = pyroute2.IPRoute()
            link_idx = ip.link_lookup(ifname=ifname)[0]
            ip.link('set', index=link_idx, state=state)
        except IndexError:
            raise exceptions.NetworkInterfaceNotFound(ifname)
        except pyroute2.NetlinkError as e:
            raise exceptions.NetlinkRuntimeError(e)

    def set_vf_guid(self, pf_ifname, vf_idx, guid):
        """Set vf administrative port and node GUID

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :param guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format
                     where x is a hexadecimal digit.
        """
        try:
            ip = pyroute2.IPRoute()
            link_idx = ip.link_lookup(ifname=pf_ifname)[0]
            ip.link('set', index=link_idx, vf={'vf': int(vf_idx),
                                               'ib_port_guid': guid})
            ip.link('set', index=link_idx,
                    vf={'vf': int(vf_idx), 'ib_node_guid': guid})
        except IndexError:
            raise exceptions.NetworkInterfaceNotFound(pf_ifname)
        except pyroute2.NetlinkError as e:
            raise exceptions.NetlinkRuntimeError(e)

    def get_vf_guid(self, pf_ifname, vf_idx):
        """Get vf administrative GUID

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :returns vf_guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format
                          where x is a hexadecimal digit.

        NOTE: while there are two GUIDs assigned per VF (port and node GUID)
        we assume they are the same and return just one value.
        """
        try:
            ip = pyroute2.IPRoute()
            link_idx = ip.link_lookup(ifname=pf_ifname)[0]
            attrs = ip.link('get', index=link_idx, ext_mask=1)[0]
        except IndexError:
            raise exceptions.NetworkInterfaceNotFound(pf_ifname)
        except pyroute2.NetlinkError as e:
            raise exceptions.NetlinkRuntimeError(e)

        vf_attr = (attrs.get_attr('IFLA_VFINFO_LIST').
                   get_attrs("IFLA_VF_INFO"))[int(vf_idx)]
        node_guid_attr = vf_attr.get_attr("IFLA_VF_IB_NODE_GUID")
        port_guid_attr = vf_attr.get_attr("IFLA_VF_IB_PORT_GUID")

        if node_guid_attr is None or port_guid_attr is None:
            # Note(adrianc) both attributes are expected to be present
            raise exceptions.NetlinkAttrNotFoundError(
                "IFLA_VF_IB_NODE_GUID, IFLA_VF_IB_PORT_GUID")

        node_guid = node_guid_attr["ib_node_guid"]
        port_guid = port_guid_attr["ib_port_guid"]

        if node_guid != port_guid:
            # Note(adrianc) both attributes are expected to be the same
            raise exceptions.NetlinkUnexpectedAttrValue(
                "port and node GUID are expected to be the same for "
                "%(netdev)s-vf%(vf_idx)s. actual: %(node_guid)s, %(port_guid)s"
                % dict(netdev=pf_ifname,
                       vf_idx=str(vf_idx),
                       node_guid=node_guid,
                       port_guid=port_guid))
        return port_guid



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_log import log as logging

from networking_mlnx.internal.sys_ops import exceptions
from networking_mlnx.internal.sys_ops import sys_ops_abs

LOG = logging.getLogger(__name__)


class OsCommand(sys_ops_abs.SystemOps):

    SYS_BASE_PATH = os.path.join(os.sep, "sys", os.sep)

    def sys_write(self, path, value):
        """Write to sysfs path (/sys/...)

        :param path: sysfs path
        :param value: value to write
        :raises InvalidSystemPathError: provided path is not under sysfs
        :raises SystemOperationError: an error occured when opening/writing
                                      to file.
        """
        if not os.path.realpath(path).startswith(OsCommand.SYS_BASE_PATH):
            raise exceptions.InvalidSystemPathError(path)
        try:
            with open(path, 'w') as fd:
                fd.write("%s" % value)
        except Exception as e:
            raise exceptions.SystemOperationError(e)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from networking_mlnx._i18n import _


interface_driver_opts = [
    cfg.StrOpt('ipoib_physical_interface',
           default="ib0",
           help=_("Name of the IPoIB root device to use with"
                  "ipoib interface driver.")),
    cfg.StrOpt('multi_interface_driver_mappings',
               default=None,
               help=_("A per physnet interface driver mapping used by "
                      "multidriver interface driver to manage the virtual "
                      "interface per physnet. a virtual network e.g vxlan "
                      "will map to the 'nil' physnet.")),
    cfg.BoolOpt('enable_multi_interface_driver_cache_maintenance',
                default=True,
                help=_("Enable periodic job to perform maintenance to the"
                       "embedded network cache for multi interface driver. "
                       "Set to true if a multi interface driver instance will "
                       "be active for an extended amount of time."))
]



# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.plugins.ml2.drivers import mech_agent
from neutron_lib.api.definitions import extra_dhcp_opt as edo_ext
from neutron_lib.api.definitions import port
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as p_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api

AGENT_TYPE_MLNX = 'Mellanox plugin agent'
VIF_TYPE_IB_HOSTDEV = 'ib_hostdev'


class MlnxMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using Mellanox eSwitch L2 agent.

    The MellanoxMechanismDriver integrates the ml2 plugin with the
    Mellanox eswitch L2 agent. Port binding with this driver requires the
    Mellanox eswitch  agent to be running on the port's host, and that agent
    to have connectivity to at least one segment of the port's
    network.
    """

    def __init__(self):
        super(MlnxMechanismDriver, self).__init__(
            agent_type=AGENT_TYPE_MLNX,
            vif_type=VIF_TYPE_IB_HOSTDEV,
            vif_details={portbindings.CAP_PORT_FILTER: False},
            supported_vnic_types=[portbindings.VNIC_DIRECT])

    def get_allowed_network_types(self, agent=None):
        return [p_constants.TYPE_LOCAL, p_constants.TYPE_FLAT,
                p_constants.TYPE_VLAN]

    def get_mappings(self, agent):
        return agent['configurations'].get('interface_mappings', {})

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        if self.check_segment_for_agent(segment, agent):
            if (segment[api.NETWORK_TYPE] in
                    (p_constants.TYPE_FLAT, p_constants.TYPE_VLAN)):
                self.vif_details['physical_network'] = segment[
                    'physical_network']
            context.set_binding(segment[api.ID],
                                self.vif_type,
                                self.vif_details)

    def _gen_client_id(self, port):
        _PREFIX = 'ff:00:00:00:00:00:02:00:00:02:c9:00:'
        _MIDDLE = ':00:00:'
        mac_address = port["mac_address"]
        mac_first = mac_address[:8]
        mac_last = mac_address[9:]
        client_id = ''.join([_PREFIX, mac_first, _MIDDLE, mac_last])
        return client_id

    def _gen_client_id_opt(self, port):
        client_id = self._gen_client_id(port)
        return [{"opt_name": edo_ext.DHCP_OPT_CLIENT_ID,
                 "opt_value": client_id}]

    def _gen_none_client_id_opt(self, port):
        updated_extra_dhcp_opts = []
        for opt in port["extra_dhcp_opts"]:
            opt["opt_value"] = None
            updated_extra_dhcp_opts.append(opt)
        return updated_extra_dhcp_opts

    def _process_port_info(self, context):
        original_port = context.original
        updated_port = context.current
        original_host_id = original_port.get("binding:host_id")
        current_host_id = updated_port.get("binding:host_id")
        # in case migration did not take place or delete port
        if original_host_id == current_host_id or current_host_id is None:
            return
        plugin = directory.get_plugin()
        if updated_port.get("extra_dhcp_opts"):
            if updated_port["extra_dhcp_opts"] == \
                    self._gen_client_id_opt(updated_port):
                return
        if not updated_port["device_owner"]:
            updated_port["extra_dhcp_opts"] = \
                    self._gen_none_client_id_opt(updated_port)
            plugin._update_extra_dhcp_opts_on_port(
                context._plugin_context,
                updated_port["id"],
                {port.RESOURCE_NAME: updated_port})
        elif ("compute" in updated_port["device_owner"] and
                updated_port["binding:vnic_type"] in ("direct", "normal")):
            updated_port["extra_dhcp_opts"] = \
                    self._gen_client_id_opt(updated_port)
            plugin._update_extra_dhcp_opts_on_port(
                context._plugin_context,
                updated_port["id"],
                {port.RESOURCE_NAME: updated_port})

    def update_port_precommit(self, context):
        self._process_port_info(context)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import os
import re

from oslo_log import log as logging

from networking_mlnx._i18n import _
from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.common import exceptions

LOG = logging.getLogger(__name__)


class pciUtils(object):

    ETH_PATH = "/sys/class/net/%(interface)s"
    ETH_DEV = ETH_PATH + "/device"
    ETH_DRIVER = ETH_DEV + "/driver"
    ETH_PORT = ETH_PATH + "/dev_id"
    INFINIBAND_PATH = 'device/infiniband'
    VENDOR_PATH = ETH_DEV + '/vendor'
    _VIRTFN_RE = re.compile(r'virtfn(?P<vf_num>\d+)')
    VFS_PATH = ETH_DEV + "/virtfn*"
    PCI_NET_PATH = ETH_DEV + "/virtfn%(vf_num)d/net"
    IB_DEV_PATH = "/sys/class/infiniband/%(ib_dev)s/device"
    VF_PCI_DEV_PATH = "/sys/bus/pci/devices/%(pf_pci_addr)s/virtfn%(vf_num)d"

    def get_vfs_info(self, pf):
        """Get VFs information

        :param pf: PF net device name
        :return: a dict containing VF info of the given PF
                 dict format example: {'04:00.3' : {'vf_num': 2,
                                                    'vf_device_type': 'MLNX5'
                                                   },
                                       ...
                                      }
        """
        vfs_info = {}
        try:
            dev_path = self.ETH_DEV % {'interface': pf}
            dev_info = os.listdir(dev_path)
            device_type = self.get_pf_device_type(pf)
            for dev_filename in dev_info:
                result = self._VIRTFN_RE.match(dev_filename)
                if result and result.group('vf_num'):
                    dev_file = os.path.join(dev_path, dev_filename)
                    vf_pci = os.readlink(dev_file).strip("./")
                    vf_num = int(result.group('vf_num'))
                    vf_device_type = device_type
                    vfs_info[vf_pci] = {'vf_num': vf_num,
                                        'vf_device_type': vf_device_type}
        except Exception as e:
            LOG.error("PCI device %s not found. %s", pf, str(e))
        LOG.info("VFs info for PF %s: %s", pf, vfs_info)
        return vfs_info

    def get_dev_attr(self, attr_path):
        try:
            fd = open(attr_path)
            return fd.readline().strip()
        except IOError:
            return

    def verify_vendor_pf(self, pf, vendor_id=constants.VENDOR):
        """Ensure PF net device PCI vendor ID equals vendor_id

        :param pf: PF netdev name
        :param vendor_id: PCI vendor ID
        :return: True if the PCI device id of the PF equals vendor_id
                 else false
        """
        vendor_path = pciUtils.VENDOR_PATH % {'interface': pf}
        if self.get_dev_attr(vendor_path) == vendor_id:
            return True
        else:
            return False

    def get_pf_device_type(self, pf):
        device_type = None
        try:
            driver_type = os.readlink(self.ETH_DRIVER
                                      % {'interface': pf})
            driver_type = os.path.basename(driver_type)
            if driver_type == constants.MLNX4_DRIVER_TYPE:
                device_type = constants.MLNX4_DEVICE_TYPE
            elif driver_type == constants.MLNX5_DRIVER_TYPE:
                device_type = constants.MLNX5_DEVICE_TYPE
            else:
                raise Exception(_('driver type %s is not supported'),
                                driver_type)
        except IOError:
            pass
        return device_type

    def is_sriov_pf(self, pf):
        """Checks if PF net dev exists and SR-IOV is enabled

        :param pf: pf netdev name
        :return: True if the device exists and has SR-IOV enabled.
        """
        vfs_path = pciUtils.VFS_PATH % {'interface': pf}
        vfs = glob.glob(vfs_path)
        if vfs:
            return True
        else:
            return

    def get_pf_mlx_dev(self, pf):
        """Get PF Infiniband device (AKA mlx device)

        :param pf: pf netdev name
        :return: pf mlx device name
        """
        dev_path = (
            os.path.join(pciUtils.ETH_PATH % {'interface': pf},
            pciUtils.INFINIBAND_PATH))
        dev_info = os.listdir(dev_path)
        return dev_info.pop()

    def get_pci_from_ib_dev(self, ib_dev):
        """Get PF PCI address from IB device

        :param ib_dev: ib device
        :return: PCI address associated with the provided IB device
        :raises: DeviceNotFoundException
        """
        pf_dev_path = pciUtils.IB_DEV_PATH % {'ib_dev': ib_dev}
        if os.path.exists(pf_dev_path):
            return os.readlink(pf_dev_path).split(os.sep)[-1]
        raise exceptions.DeviceNotFoundException(
            "PCI address was not found for IB device %s" % ib_dev)

    def get_vf_from_vf_idx(self, pf_pci, vf_idx):
        """Get VF PCI address from PF PCI address and VF index

        :param pf_pci: PF PCI address D:B:D.F format
        :param vf_idx: VF index
        :return: VF PCI address
        :raises: DeviceNotFoundException
        """
        vf_dev_path = pciUtils.VF_PCI_DEV_PATH % {
            'pf_pci_addr': pf_pci, 'vf_num': vf_idx}
        if os.path.exists(vf_dev_path):
            return os.readlink(vf_dev_path).split(os.sep)[-1]
        raise exceptions.DeviceNotFoundException(
            "VF PCI address not found for PF %s VF index %d " % (
                pf_pci, vf_idx))

    def get_eth_port(self, dev):
        """Get network device Port number

        :param dev: netdev name
        :return: HCA port number
        """
        port_path = pciUtils.ETH_PORT % {'interface': dev}
        try:
            with open(port_path) as f:
                dev_id = int(f.read(), 0)
                return dev_id + 1
        except IOError:
            return

    def is_assigned_vf(self, pf_name, vf_index):
        """Check if VF is assigned.

       Checks if a given vf index of a given device name is assigned
       by checking the relevant path in the system:
       VF is assigned if PCI_PATH does not exist.
       @param pf_name: pf network device name
       @param vf_index: vf index
        """
        if not self.is_sriov_pf(pf_name):
            # If the root PCI path does not exist or has no VFs then
            # the VF cannot actually have been allocated and there is
            # no way we can manage it.
            return False

        path = self.PCI_NET_PATH % {'interface': pf_name, 'vf_num': vf_index}

        return not os.path.exists(path)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MlxException(Exception):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return 'MlxException: %s' % self.message


class DeviceNotFoundException(MlxException):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return 'PCI device not found: %s' % self.message


class InvalidGUIDFormatException(MlxException):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return 'Invalid GUID format: %s' % self.message



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import lockutils
from oslo_log import log

from networking_mlnx.linux.interface_drivers import constants

LOG = log.getLogger(__name__)


class SimpleCache(object):
    """A simple object cache"""

    def __init__(self):
        self.object_cache = {}

    def __contains__(self, id):
        return id in self.object_cache

    def get(self, id):
        """Get object from cache

        :param id: object id
        :return: object or None if id is not in cache
        """
        return self.object_cache.get(id)

    def get_all(self):
        """Get all object from cache

        :return: list of objects
        """
        return list(self.object_cache.values())

    def put(self, id, obj):
        """Add object to cache

        :param id: object id
        :param obj: object to add
        """
        self.object_cache[id] = obj

    def remove(self, id):
        """Remove object from cache

        :param id: object id
        """
        if id in self:
            del self.object_cache[id]

    def clear(self):
        """Clear the cache"""
        self.object_cache = {}


class NetworkCache(SimpleCache):
    """A network cache that retrieves missing entries
    via callback.
    """

    def __init__(self, get_networks_cb, fields):
        """__init__

        :param get_networks_cb: a callback to get network object
               the callback should accept filters and fields parameters.
               e.g: def get_networks(filters, fields)
        :param fields: a list of fields to get for a network object
                       used as an argument for get_networks_cb when getting
                       new network objects. None value will fetch all available
                       fields.
        """
        super(NetworkCache, self).__init__()
        self.get_networks_cb = get_networks_cb
        self.network_fields = fields
        if self.network_fields and 'id' not in self.network_fields:
            self.network_fields.append('id')

    def _get_no_callback(self, network_id):
        """Get network from cache
        in case of a cache miss, return None.

        :param network_id: network id
        :return: network dict or None
        """
        return super(NetworkCache, self).get(network_id)

    def get(self, network_id):
        """Get network from cache
        in case of a cache miss, attempt to get network from callback.

        :param network_id: network id
        :return: network dict
        """
        net = self._get_no_callback(network_id)
        if net is None:
            LOG.debug("Network %s not in cache, fetching via callback.",
                      network_id)
            net = self.get_networks_cb(filters={'id': [network_id]},
                                       fields=self.network_fields)[0]
            self.put(net['id'], net)
        return net

    def refresh(self):
        """Refresh network cache"""
        nets = self.get_networks_cb(filters={constants.ADMIN_STATE_UP: [True]},
                                    fields=self.network_fields)
        self.clear()
        for net in nets:
            self.put(net['id'], net)

    def remove_stale_networks(self):
        """Remove networks that no longer exist or are inactive from cache

        This method is intended to be used as a periodic job to prevent the
        cache from containing too many stale entries.
        """
        nets = self.get_networks_cb(filters={constants.ADMIN_STATE_UP: [True]},
                                    fields=['id'])
        current_net_ids = set([net['id'] for net in nets])
        cached_net_ids = set([net['id'] for net in self.get_all()])
        ids_to_remove = cached_net_ids - current_net_ids
        if ids_to_remove:
            LOG.debug("Removing stale networks from cache: %s", ids_to_remove)
            for id in ids_to_remove:
                self.remove(id)


class SafeNetworkCache(NetworkCache):
    """Thread safe implementation of NetworkCache
    Which basically wraps operations with a ReaderWriter Lock
    """

    def __init__(self, get_networks_cb, fields):
        """__init__

        :param get_networks_cb: a callback to get network object
               the callback should accept filters and fields parameters.
               e.g: def get_networks(filters, fields)
        :param fields: a list of fields to get for a network object
                       used as an argument for get_networks_cb when getting
                       new network objects. None value will fetch all available
                       fields.
        """
        super(SafeNetworkCache, self).__init__(get_networks_cb, fields)
        self.__rw_lock = lockutils.ReaderWriterLock()

    def get(self, id):
        """Get network from cache
        in case of a cache miss, attempt to get network from callback.

        :param id: network id
        :return: network dict
        """
        with self.__rw_lock.read_lock():
            net = self._get_no_callback(id)

        if net is None:
            LOG.debug("Network %s not in cache, fetching via callback.", id)
            net = self.get_networks_cb(filters={'id': [id]},
                                       fields=self.network_fields)[0]
            # Put operation is already protected by a writer lock
            self.put(net['id'], net)
        return net

    def get_all(self):
        """Get all object from cache

        :return: list of objects
        """
        with self.__rw_lock.read_lock():
            return super(SafeNetworkCache, self).get_all()

    def put(self, id, obj):
        """Add object to cache

        :param id: object id
        :param obj: object to add
        """
        with self.__rw_lock.write_lock():
            return super(SafeNetworkCache, self).put(id, obj)

    def remove(self, id):
        """Remove object from cache

        :param id: object id
        """
        with self.__rw_lock.write_lock():
            return super(SafeNetworkCache, self).remove(id)

    def clear(self):
        """Clear the cache"""
        with self.__rw_lock.write_lock():
            return super(SafeNetworkCache, self).clear()

    def refresh(self):
        """Refresh network cache"""
        with self.__rw_lock.write_lock():
            return super(SafeNetworkCache, self).refresh()



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

VENDOR = '0x15b3'
VIF_TYPE_HOSTDEV = 'ib_hostdev'

UNTAGGED_VLAN_ID = 4095

INVALID_MAC = '00:00:00:00:00:00'

# MLNX4
MLNX4_ADMIN_GUID_PATH = "/sys/class/infiniband/%s/iov/ports/%s/admin_guids/%s"
MLNX4_GUID_INDEX_PATH = "/sys/class/infiniband/%s/iov/%s/ports/%s/gid_idx/0"
MLNX4_PKEY_INDEX_PATH = "/sys/class/infiniband/%s/iov/%s/ports/%s/pkey_idx/%s"

# MLNX5
MLNX5_GUID_NODE_PATH = ('/sys/class/infiniband/%(module)s/device/sriov/'
                      '%(vf_num)d/node')
MLNX5_GUID_PORT_PATH = ('/sys/class/infiniband/%(module)s/device/sriov/'
                      '%(vf_num)d/port')
UNBIND_PATH = '/sys/bus/pci/drivers/mlx5_core/unbind'
BIND_PATH = '/sys/bus/pci/drivers/mlx5_core/bind'

MLNX4_INVALID_GUID = 'ffffffffffffffff'
MLNX5_INVALID_GUID = 'ff:ff:ff:ff:ff:ff:ff:ff'

CONN_URL = '%(transport)s://%(addr)s:%(port)s'

MLNX4_DRIVER_TYPE = 'mlx4_core'
MLNX5_DRIVER_TYPE = 'mlx5_core'

MLNX4_DEVICE_TYPE = 'MLNX4'
MLNX5_DEVICE_TYPE = 'MLNX5'

SOCKET_OS_PORT = '60001'
SOCKET_OS_TRANSPORT = 'tcp'
SOCKET_OS_ADDR = '0.0.0.0'



#!/usr/bin/python
# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys

from networking_mlnx.eswitchd.cli import conn_utils
from networking_mlnx.eswitchd.cli import exceptions

action = sys.argv[1]
client = conn_utils.ConnUtil()


def pprint_table(out, table):
    """Prints out a table of data, padded for alignment

    @param out: Output stream (file-like object)
    @param table: The table to print. A list of lists.
    Each row must have the same number of columns.
    """

    def get_max_width(table, index):
        """Get the maximum width of the given column index"""
        return max([len(str(row[index])) for row in table])

    col_paddings = []

    for i in range(len(table[0])):
        col_paddings.append(get_max_width(table, i))

    for row in table:
        # left col
        print(row[0].ljust(col_paddings[0] + 1), file=out)
        # rest of the cols
        for i in range(1, len(row)):
            col = str(row[i]).rjust(col_paddings[i] + 2)
            print(col, file=out)
        print(file=out)


def main():
    if action == 'get-tables':
        fabric = sys.argv[2]
        try:
            result = client.get_tables(fabric)
            for fabric, tables in result.items():
                print("FABRIC = %s" % fabric)
                print("========================")
                for table, data in tables.items():
                    print("TABLE: %s" % table)
                    pprint_table(sys.stdout, data)
                    print("========================")
        except exceptions.MlxException as e:
            sys.stderr.write("Error in get-tables command")
            sys.stderr.write(e.message)
            sys.exit(1)
        sys.exit(0)


if __name__ == '__main__':
    main()



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class NetDevOperations(object):

    @abc.abstractmethod
    def set_vf_admin_state(self, pf_ifname, vf_idx, state):
        """Set vf admin state

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :param state: desired admin state as defined in
                      networking_mlnx.internal.netdev_ops.constants
        """
        pass

    @abc.abstractmethod
    def set_link_state(self, ifname, state):
        """Set net device link state

        :param ifname: netdev name
        :param state: desired link state as defined in
                      networking_mlnx.internal.netdev_ops.constants
        """

    @abc.abstractmethod
    def set_vf_guid(self, pf_ifname, vf_idx, guid):
        """Set vf administrative port and node GUID

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :param guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format
                     where x is a hexadecimal digit.
        """

    @abc.abstractmethod
    def get_vf_guid(self, pf_ifname, vf_idx):
        """Get vf administrative GUID

        :param pf_ifname: pf netdev name
        :param vf_idx: vf index
        :returns guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format
                          where x is a hexadecimal digit.
        """



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_mlnx.eswitchd.common import exceptions
from networking_mlnx.eswitchd import eswitch_manager as esm
from networking_mlnx.internal.netdev_ops import constants as netdev_const
from networking_mlnx.tests import base


class TestCreateESwitchManager(base.TestCase):
    SCANNED_DEVICES = [('0000:06:00.1', 0),
                       ('0000:06:00.2', 1),
                       ('0000:06:00.3', 2)]

    @staticmethod
    def cleanup():
        if hasattr(esm.ESwitchManager, '_instance'):
            del esm.ESwitchManager._instance

    def test_create_eswitch_mgr_fail_vendor(self):
        device_mappings = [('physnet1', 'p6p1')]
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils.pciUtils."
                        "is_sriov_pf",
                        return_value=True), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils.pciUtils."
                           "verify_vendor_pf", return_value=False):
            eswitch_mgr = esm.ESwitchManager()
            self.addCleanup(self.cleanup)
            self.assertRaises(SystemExit,
                              eswitch_mgr.discover_devices,
                              device_mappings)

    def test_create_eswitch_mgr_fail_pf(self):
        device_mappings = [('physnet1', 'p6p1')]
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                        "pciUtils.is_sriov_pf", return_value=False), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                           "pciUtils.verify_vendor_pf", return_value=True):
            eswitch_mgr = esm.ESwitchManager()
            self.addCleanup(self.cleanup)
            self.assertRaises(SystemExit,
                              eswitch_mgr.discover_devices,
                              device_mappings)

    def test_create_eswitch_mgr_ok(self):
        device_mappings = [('physnet1', 'p6p1')]
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                        "pciUtils.is_sriov_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                           "pciUtils.verify_vendor_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.eswitch_manager."
                           "EmbSwitch"):
            eswitch_mgr = esm.ESwitchManager()
            self.addCleanup(self.cleanup)
            eswitch_mgr.discover_devices(device_mappings)


class TestESwitchManagerApi(base.TestCase):
    SCANNED_DEVICES = [('0000:06:00.1', 0),
                       ('0000:06:00.2', 1),
                       ('0000:06:00.3', 2)]

    VF_INFO_OUT = {'0000:06:00.1': {'vf_num': '0', 'vf_device_type': 'MLNX5'},
                   '0000:06:00.2': {'vf_num': '1', 'vf_device_type': 'MLNX5'},
                   '0000:06:00.3': {'vf_num': '2', 'vf_device_type': 'MLNX5'}}

    ASSIGNED_MAC = '00:00:00:00:00:66'
    PCI_SLOT = '0000:06:00.1'
    PCI_IDX = '0'
    WRONG_MAC = '00:00:00:00:00:67'
    WRONG_PCI = "0000:06:00.6"

    def setUp(self):
        super(TestESwitchManagerApi, self).setUp()
        device_mappings = [('physnet1', 'p6p1')]
        self.eswitch_mgr = esm.ESwitchManager()
        self.addCleanup(self.cleanup)
        self._set_eswitch_manager(self.eswitch_mgr, device_mappings)

    @staticmethod
    def cleanup():
        if hasattr(esm.ESwitchManager, '_instance'):
            del esm.ESwitchManager._instance

    def _set_eswitch_manager(self, eswitch_mgr, device_mappings):
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                        "pciUtils.is_sriov_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils.pciUtils."
                           "verify_vendor_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.eswitch_manager."
                "PfPciDeviceWrapper") as mock_dev_wrapper_cls:
            self.mock_dev_wrapper = mock.Mock()
            mock_dev_wrapper_cls.return_value = self.mock_dev_wrapper
            self.mock_dev_wrapper.get_vfs_info.return_value = self.VF_INFO_OUT
            eswitch_mgr.discover_devices(device_mappings)

    def test_discover_devices_with_device(self):
        device_mappings = [('physnet1', 'p6p1'), ('physnet1', 'p6p2')]
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                        "pciUtils.is_sriov_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils.pciUtils."
                           "verify_vendor_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.eswitch_manager."
                           "ESwitchManager._create_emb_switch") as emb_switch:
            self.eswitch_mgr.discover_devices(device_mappings)
            self.assertTrue(emb_switch.called)

    def test_get_assigned_devices_info(self):
        with mock.patch("networking_mlnx.eswitchd.eswitch_manager.EmbSwitch."
                        "get_devices_info",
                        return_value=[(self.ASSIGNED_MAC, self.PCI_SLOT)]):
            result = self.eswitch_mgr.get_assigned_devices_info()
            self.assertIn(self.ASSIGNED_MAC, list(result)[0])
            self.assertIn(self.PCI_SLOT, list(result)[0])

    def test_get_assigned_devices_info_multiple_nics_for_physnet(self):
        device_mappings = [('physnet1', 'p6p1'), ('physnet1', 'p6p2')]
        devices_info = {
            'p6p1': [(self.ASSIGNED_MAC, self.PCI_SLOT)],
            'p6p2': [(self.WRONG_MAC, self.WRONG_PCI)],
        }

        def my_get_devices_info(self, assigned_only):
            return devices_info[self.dev_name]

        self._set_eswitch_manager(self.eswitch_mgr, device_mappings)

        with mock.patch("networking_mlnx.eswitchd.eswitch_manager.EmbSwitch."
                        "get_devices_info",
                        side_effect=my_get_devices_info,
                        autospec=True):
            result = self.eswitch_mgr.get_assigned_devices_info()
            self.assertIn(devices_info['p6p1'][0], list(result))
            self.assertIn(devices_info['p6p2'][0], list(result))

    def test_get_device_admin_state_up(self):
        self.mock_dev_wrapper.get_vf_state.return_value = \
            netdev_const.ADMIN_STATE_UP
        result = self.eswitch_mgr.get_device_admin_state(self.PCI_SLOT)
        self.assertEqual(netdev_const.ADMIN_STATE_UP, result)

    def test_get_device_admin_state_down(self):
        self.mock_dev_wrapper.get_vf_state.return_value = \
            netdev_const.ADMIN_STATE_DOWN
        result = self.eswitch_mgr.get_device_admin_state(self.PCI_SLOT)
        self.assertEqual(netdev_const.ADMIN_STATE_DOWN, result)

    def test_get_device_admin_state_auto(self):
        self.mock_dev_wrapper.get_vf_state.return_value = \
            netdev_const.ADMIN_STATE_AUTO
        result = self.eswitch_mgr.get_device_admin_state(self.PCI_SLOT)
        self.assertEqual(netdev_const.ADMIN_STATE_AUTO, result)

    def test_set_device_admin_state(self):
        self.eswitch_mgr.set_device_admin_state(self.PCI_SLOT,
                                                netdev_const.ADMIN_STATE_UP)
        self.mock_dev_wrapper.set_vf_state.assert_called_with(
            self.PCI_IDX, netdev_const.ADMIN_STATE_UP)

    def _mock_device_exists(self, pci_slot, mac_address, expected_result):
        with mock.patch("neutron.plugins.ml2.drivers.mech_sriov.agent."
                        "eswitch_manager.EmbSwitch.get_pci_device",
                        return_value=self.ASSIGNED_MAC):
            result = self.eswitch_mgr.device_exists(mac_address,
                                                    pci_slot)
            self.assertEqual(expected_result, result)

    def test_device_exists_true(self):
        self.assertTrue(self.eswitch_mgr.device_exists(self.PCI_SLOT))

    def test_device_exists_false(self):
        self.assertFalse(self.eswitch_mgr.device_exists(self.WRONG_PCI))

    def test_process_emb_switch(self):
        device_mappings = [('physnet1', 'p6p1'), ('physnet1', 'p6p2')]
        phys_net = 'physnet1'
        dev_name = 'p6p1'
        self._set_eswitch_manager(self.eswitch_mgr, device_mappings)
        with mock.patch("networking_mlnx.eswitchd.utils.pci_utils."
                        "pciUtils.is_sriov_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.utils.pci_utils.pciUtils."
                           "verify_vendor_pf", return_value=True), \
                mock.patch("networking_mlnx.eswitchd.eswitch_manager."
                           "ESwitchManager._create_emb_switch") as emb_switch:
            self.eswitch_mgr._process_emb_switch_map(phys_net,
                                                     dev_name)
            self.assertFalse(emb_switch.called)

    def test_create_emb_switch(self):
        DEVICES = [('0000:04:00.1', 0),
                   ('0000:04:00.2', 1)]
        with mock.patch("networking_mlnx.eswitchd."
                        "eswitch_manager.EmbSwitch") as mock_esw:
            esw_inst = mock.Mock()
            mock_esw.return_value = esw_inst
            esw_inst.get_pci_slot_list.return_value = [
                slot for slot, idx in DEVICES]
            physnet = 'test_create_emb_switch'
            self.assertNotIn(physnet, self.eswitch_mgr.emb_switches_map)
            # create a new embedded switch
            self.eswitch_mgr._create_emb_switch(physnet, 'dev1')
            self.assertIn(physnet, self.eswitch_mgr.emb_switches_map)
            for slot, _ in DEVICES:
                self.assertIn(slot, self.eswitch_mgr.pci_slot_map)

    def test_create_emb_switch_zero_vfs(self):
        with mock.patch("networking_mlnx.eswitchd."
                        "eswitch_manager.EmbSwitch") as mock_esw:
            esw_inst = mock.Mock()
            mock_esw.return_value = esw_inst
            esw_inst.get_pci_slot_list.return_value = []
            physnet = 'test_create_emb_switch'
            self.eswitch_mgr.emb_switches_map = {}
            self.eswitch_mgr.pci_slot_map = {}
            # create a new embedded switch
            self.eswitch_mgr._create_emb_switch(physnet, 'dev1')
            self.assertIn(physnet, self.eswitch_mgr.emb_switches_map)
            self.assertFalse(self.eswitch_mgr.pci_slot_map)


class TestEmbSwitch(base.TestCase):
    DEV_NAME = "eth2"
    PHYS_NET = "default"
    ASSIGNED_MAC = '00:00:00:00:00:66'
    PCI_SLOT = "0000:06:00.1"
    WRONG_PCI_SLOT = "0000:06:00.4"
    VF_TO_MAC_MAPPING = {'0': '00:00:00:00:00:11',
                         '1': '00:00:00:00:00:22',
                         '2': '00:00:00:00:00:33'}
    VF_INFO_OUT = {'0000:06:00.1': {'vf_num': '0', 'vf_device_type': 'MLNX5'},
                   '0000:06:00.2': {'vf_num': '1', 'vf_device_type': 'MLNX5'},
                   '0000:06:00.3': {'vf_num': '2', 'vf_device_type': 'MLNX5'}}
    EXPECTED_DEVICE_INFO = [('00:00:00:00:00:11', '0000:06:00.1'),
                            ('00:00:00:00:00:22', '0000:06:00.2'),
                            ('00:00:00:00:00:33', '0000:06:00.3')]

    def setUp(self):
        super(TestEmbSwitch, self).setUp()
        with mock.patch("networking_mlnx.eswitchd.eswitch_manager."
                        "PfPciDeviceWrapper") as mock_dev_wrapper_cls:
            self.mock_dev_wrapper = mock.Mock()
            mock_dev_wrapper_cls.return_value = self.mock_dev_wrapper
            self.mock_dev_wrapper.get_vfs_info.return_value = self.VF_INFO_OUT
            self.emb_switch = esm.EmbSwitch(self.DEV_NAME)

    def test_get_devices_info_assigned_only_true(self):
        def my_is_assigned_vf(vf_idx):
            return vf_idx != 0
        del self.VF_TO_MAC_MAPPING['0']
        self.mock_dev_wrapper.get_vfs_macs_ib.return_value = \
            self.VF_TO_MAC_MAPPING
        self.mock_dev_wrapper.is_assigned_vf.side_effect = my_is_assigned_vf

        result = self.emb_switch.get_devices_info(True)
        self.assertNotIn(('00:00:00:00:00:11', '0000:06:00.1'), result)
        self.assertIn(('00:00:00:00:00:22', '0000:06:00.2'), result)
        self.assertIn(('00:00:00:00:00:33', '0000:06:00.3'), result)
        self.assertEqual(2, len(result))

    def test_get_devices_info_assigned_only_false(self):
        self.mock_dev_wrapper.get_vfs_macs_ib.return_value = \
            self.VF_TO_MAC_MAPPING
        self.mock_dev_wrapper.is_assigned_vf.return_value = False
        result = self.emb_switch.get_devices_info()
        self.assertEqual(set(self.EXPECTED_DEVICE_INFO), set(result))

    def test_get_device_state_ok(self):
        self.mock_dev_wrapper.get_vf_state.return_value = \
            netdev_const.ADMIN_STATE_DOWN
        result = self.emb_switch.get_device_admin_state(self.PCI_SLOT)
        self.assertEqual(netdev_const.ADMIN_STATE_DOWN, result)

    def test_get_device_state_fail(self):
        self.mock_dev_wrapper.get_vf_state.return_value = \
            netdev_const.ADMIN_STATE_DOWN
        self.assertRaises(exceptions.DeviceNotFoundException,
                          self.emb_switch.get_device_admin_state,
                          self.WRONG_PCI_SLOT)

    def test_set_device_admin_state_ok(self):
        self.emb_switch.set_device_admin_state(
            self.PCI_SLOT, netdev_const.ADMIN_STATE_UP)
        self.assertTrue(self.mock_dev_wrapper.set_vf_state.called)

    def test_set_device_admin_state_fail(self):
        self.assertRaises(exceptions.DeviceNotFoundException,
                          self.emb_switch.set_device_admin_state,
                          self.WRONG_PCI_SLOT, netdev_const.ADMIN_STATE_UP)
        self.assertFalse(self.mock_dev_wrapper.set_vf_state.called)

    def test_get_device_ib_mac(self):
        self.mock_dev_wrapper.is_assigned_vf.return_value = True
        self.mock_dev_wrapper.get_vfs_macs_ib.return_value = {
            '0': self.ASSIGNED_MAC}
        result = self.emb_switch.get_device_ib_mac(self.PCI_SLOT)
        self.assertEqual(self.ASSIGNED_MAC, result)

    def test_get_device_ib_mac_unassigned(self):
        self.mock_dev_wrapper.is_assigned_vf.return_value = False
        result = self.emb_switch.get_device_ib_mac(self.PCI_SLOT)
        self.assertIsNone(result)

    def test_get_pci_list(self):
        result = self.emb_switch.get_pci_slot_list()
        self.assertEqual(set(self.VF_INFO_OUT.keys()), set(result))

    def test_set_device_ib_mac(self):
        self.emb_switch.set_device_ib_mac(self.PCI_SLOT, self.ASSIGNED_MAC)
        self.assertTrue(self.mock_dev_wrapper.config_vf_mac_address.called)

    def test_set_device_ib_mac_invalid_device(self):
        self.assertRaises(exceptions.DeviceNotFoundException,
                          self.emb_switch.set_device_ib_mac,
                          self.WRONG_PCI_SLOT,
                          self.ASSIGNED_MAC)
        self.assertFalse(self.mock_dev_wrapper.config_vf_mac_address.called)

    def test_config_vlan_ib(self):
        self.emb_switch.config_vlan_ib(self.PCI_SLOT, 12)
        self.assertTrue(self.mock_dev_wrapper.config_vlan_ib.called)


class TestPfPciDeviceWrapper(base.TestCase):
    # TODO(adrianc): add Tests
    pass



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_mlnx.internal.netdev_ops import impl_pyroute2
from networking_mlnx import privsep

# for now we have only one implementation
__netdev_ops = impl_pyroute2.IpCommand()


@privsep.admin_pctx.entrypoint
def set_vf_admin_state(pf_ifname, vf_idx, state):
    """Set vf admin state

    :param pf_ifname: pf netdev name
    :param vf_idx: vf index
    :param state: desired admin state as defined in
                  networking_mlnx.internal.netdev_ops.constants
    """
    __netdev_ops.set_vf_admin_state(pf_ifname, vf_idx, state)


@privsep.admin_pctx.entrypoint
def set_link_state(ifname, state):
    """Set net device link state

    :param ifname: netdev name
    :param state: desired link state as defined in
                  networking_mlnx.internal.netdev_ops.constants
    """
    __netdev_ops.set_link_state(ifname, state)


@privsep.admin_pctx.entrypoint
def set_vf_guid(pf_ifname, vf_idx, guid):
    """Set vf administrative port and node GUID

    :param pf_ifname: pf netdev name
    :param vf_idx: vf index
    :param guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format
                 where x is a hexadecimal digit.
    """
    __netdev_ops.set_vf_guid(pf_ifname, vf_idx, guid)


def get_vf_guid(pf_ifname, vf_idx):
    """Get vf administrative GUID

    :param pf_ifname: pf netdev name
    :param vf_idx: vf index
    :returns vf_guid: 64bit guid str in xx:xx:xx:xx:xx:xx:xx:xx format where
                      x is a hexadecimal digit.

    NOTE: This operation does not need elevated privileges.
    """
    return __netdev_ops.get_vf_guid(pf_ifname, vf_idx)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import six

from networking_mlnx._i18n import _
from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.common import exceptions
from networking_mlnx.eswitchd.utils import ib_utils
from networking_mlnx.eswitchd.utils import pci_utils
from networking_mlnx.internal.netdev_ops import api as net_dev_api
from networking_mlnx.internal.netdev_ops import constants as netdev_consts

LOG = logging.getLogger(__name__)


class PfPciDeviceWrapper(object):
    """A wrapper classs for PF PCI device used to perform Operations
    on device.

    """
    def __init__(self, dev_name):
        """Constructor

        :param dev_name: pf netdev name
        """
        # TODO(adrianc): these two (pci and ib utils) need to be sorted out
        # as the responsibility of each class is not really well defined.
        self.pci_utils = pci_utils.pciUtils()
        self.ib_utils = ib_utils.IbUtils()
        self.net_dev_name = dev_name
        self.mlx_dev_name = self.pci_utils.get_pf_mlx_dev(dev_name)
        # NOTE(adrianc): hca_port is required for ConnectX-3 family devices
        self.hca_port = self.pci_utils.get_eth_port(dev_name)
        self.device_type = self.pci_utils.get_pf_device_type(dev_name)

    def get_vfs_info(self):
        return self.pci_utils.get_vfs_info(self.net_dev_name)

    def is_assigned_vf(self, vf_num):
        return self.pci_utils.is_assigned_vf(self.net_dev_name, vf_num)

    def get_vfs_macs_ib(self, vf_idxs):
        return self.ib_utils.get_vfs_macs_ib(
            self.net_dev_name, self.mlx_dev_name, self.hca_port, vf_idxs,
            self.device_type)

    def get_vf_state(self, vf_index):
        # TODO(adrianc) implement
        raise NotImplementedError(_("get_vf_state() not implemented."))

    def set_vf_state(self, vf_index, state):
        """Set VF admin state
        """
        if self.device_type == constants.MLNX4_DEVICE_TYPE:
            LOG.warning(
                "Changing VF admin state for MLX4 device is not supported.")
        else:
            net_dev_api.set_vf_admin_state(self.net_dev_name, vf_index, state)

    def config_port_up(self):
        """Set PF netdev link state to UP
        """
        net_dev_api.set_link_state(self.net_dev_name,
                                   netdev_consts.LINK_STATE_UP)

    def config_vf_mac_address(self, vf_idx, vf_pci_slot, mac=None):
        """Configure VF mac address

        :param vf_idx: VF index
        :param vf_pci_slot: VF PCI address
        :param mac: mac address, None value means mac address cleanup
        """
        self.ib_utils.config_vf_mac_address(
            self.net_dev_name, self.mlx_dev_name, vf_idx, vf_pci_slot,
            self.device_type, self.hca_port, mac)

    def config_vlan_ib(self, pci_slot, vlan):
        """Set VF IB VLAN (Set Infiniband PKEY for VF, ConnectX3 Only)

        :param pci_slot: VF PCI slot address
        :param vlan: VLAN ID
        """
        return self.ib_utils.config_vlan_ib(
            self.mlx_dev_name, pci_slot, self.device_type, self.hca_port,
            vlan)


class EmbSwitch(object):
    """Class to manage logical embedded switch entity.

    Embedded Switch object is logical entity representing all VFs
    connected to a physical network
    Each physical network is mapped to PF network device interface,
    meaning all its VF.
    :ivar pci_slot_map: dictionary for mapping each pci slot to vf index
    :ivar pci_dev_wrapper: pci device wrapper
    """

    def __init__(self, dev_name):
        """Constructor

        :param dev_name: physical network device name
        """
        self.pf_dev_wrapper = PfPciDeviceWrapper(dev_name)
        self.dev_name = dev_name
        self.vf_pci_to_idx = {}
        vfs_info = self.pf_dev_wrapper.get_vfs_info()
        for pci_slot, vf_info in six.iteritems(vfs_info):
            self.vf_pci_to_idx[pci_slot] = vf_info['vf_num']

    def get_pci_slot_list(self):
        """Get list of VF addresses."""
        return self.vf_pci_to_idx.keys()

    def get_devices_info(self, assigned_only=False):
        """Get Virtual Functions mac and pci slot
        information

        :param assigned_only: Retrieve only assigned devices
        :return: list of VF pair (mac address, pci slot)
        """
        vf_to_pci_slot_mapping = {}
        assigned_devices_info = []
        for pci_slot, vf_index in six.iteritems(self.vf_pci_to_idx):
            if (assigned_only and
                    not self.pf_dev_wrapper.is_assigned_vf(vf_index)):
                continue
            vf_to_pci_slot_mapping[vf_index] = pci_slot

        if vf_to_pci_slot_mapping:
            vf_to_mac_mapping = self.pf_dev_wrapper.get_vfs_macs_ib(
                list(vf_to_pci_slot_mapping.keys()))
            for vf_index, mac in six.iteritems(vf_to_mac_mapping):
                pci_slot = vf_to_pci_slot_mapping[vf_index]
                assigned_devices_info.append((mac, pci_slot))
        return assigned_devices_info

    def get_device_admin_state(self, pci_slot):
        """Get device admin state.

        :param pci_slot: Virtual Function address
        :return: device admin state, one of: netdev_consts.ADMIN_STATE_*
        """
        vf_index = self._get_vf_index(pci_slot)
        return self.pf_dev_wrapper.get_vf_state(vf_index)

    def set_device_admin_state(self, pci_slot, admin_state):
        """Set device admin state.

        :param pci_slot: Virtual Function address
        :param admin_state: device admin state,
               one of: netdev_consts.ADMIN_STATE_*
        """
        vf_index = self._get_vf_index(pci_slot)
        return self.pf_dev_wrapper.set_vf_state(vf_index, admin_state)

    def _get_vf_index(self, pci_slot):
        vf_index = self.vf_pci_to_idx.get(pci_slot)
        if vf_index is None:
            raise exceptions.DeviceNotFoundException(
                _("Cannot find vf index for pci slot %s") % pci_slot)
        return vf_index

    def get_device_ib_mac(self, pci_slot):
        """Get IB mac address for given Virtual Function address

        :param pci_slot: VF pci address
        :return: MAC address of virtual function
        """
        vf_index = self._get_vf_index(pci_slot)
        mac = None
        if self.pf_dev_wrapper.is_assigned_vf(vf_index):
            macs = self.pf_dev_wrapper.get_vfs_macs_ib([vf_index])
            mac = macs.get(vf_index)
        return mac

    def set_device_ib_mac(self, pci_slot, mac=None):
        """Set IB mac address for given virtual function address

        :param pci_slot: VF pci address
        :param mac: MAC address for the VF, if None then the device
                    mac will be invalidated.

        Note:
        When invalidating MAC address (mac=None), VF admin state will change
        to disable.
        """
        vf_index = self._get_vf_index(pci_slot)
        self.pf_dev_wrapper.config_vf_mac_address(vf_index, pci_slot, mac)

    def config_vlan_ib(self, pci_slot, vlan):
        """Set VF IB VLAN (Set Infiniband PKEY for VF, ConnectX3 Only)

        :param pci_slot: VF PCI slot address
        :param vlan: VLAN ID
        """
        self.pf_dev_wrapper.config_vlan_ib(pci_slot, vlan)


class ESwitchManager(object):
    """Manages logical Embedded Switch entities for physical network."""

    def __new__(cls):
        # make it a singleton
        if not hasattr(cls, '_instance'):
            cls._instance = super(ESwitchManager, cls).__new__(cls)
            cls.emb_switches_map = {}
            cls.pci_slot_map = {}
        return cls._instance

    def __init__(self):
        super(ESwitchManager, self).__init__()
        self.pci_utils = pci_utils.pciUtils()

    def device_exists(self, pci_slot):
        """Verify if device exists.

        Check if VF pci slot exists
        :param pci_slot: VF address
        """
        embedded_switch = self._get_emb_eswitch(pci_slot)
        if embedded_switch:
            return True
        return False

    def get_assigned_devices_info(self, phys_nets=None):
        """Get all assigned devices.

        Get all assigned devices belongs to given embedded switch
        :param phys_net: list of physical networks, if none get assigned
                         devices from all physical networks
        :return: set of assigned VFs (mac address, pci slot) pairs
        """
        if phys_nets:
            eswitch_objects = set()
            for net in phys_nets:
                emb_sw = self.emb_switches_map.get(net)
                if emb_sw:
                    eswitch_objects.add(emb_sw)
        else:
            eswitch_objects = set()
            for eswitch_list in self.emb_switches_map.values():
                eswitch_objects |= set(eswitch_list)
        assigned_devices = set()
        for embedded_switch in eswitch_objects:
            for device in embedded_switch.get_devices_info(True):
                assigned_devices.add(device)
        return assigned_devices

    def get_device_from_mac(self, mac, phys_net=None):
        """Get PCI VF address from Mac address

        :param mac: MAC address
        :param phys_net: The physical network name the mac address resides in
        :return: VF PCI address that has `mac` assigned as its L2 address
                 or None if no matching device found.
        """
        if phys_net:
            eswitch_objects = self.emb_switches_map.get(phys_net, set())
        else:
            eswitch_objects = set()
            for eswitch_list in self.emb_switches_map.values():
                eswitch_objects |= set(eswitch_list)

        for embedded_switch in eswitch_objects:
            for mac_addr, slot in embedded_switch.get_devices_info():
                if mac == mac_addr:
                    # TODO(adrianc) decide if we want to deal with same MAC in
                    # multiple fabrics.
                    return slot
        return None

    def get_device_admin_state(self, pci_slot):
        """Get device state.

        Get the device state (up/enable, down/disable, or auto)
        :param pci_slot: VF PCI slot
        :return: device admin state, one of: netdev_consts.ADMIN_STATE_*
        """
        embedded_switch = self._get_emb_eswitch(pci_slot)
        if embedded_switch:
            return embedded_switch.get_device_admin_state(pci_slot)
        return netdev_consts.ADMIN_STATE_DOWN

    def set_device_admin_state(self, pci_slot, admin_state):
        """Set device admin state

        :param pci_slot: pci slot
        :param admin_state: device admin state,
               one of: netdev_consts.ADMIN_STATE_*
        """
        embedded_switch = self._get_emb_eswitch(pci_slot)
        if embedded_switch:
            embedded_switch.set_device_admin_state(pci_slot, admin_state)

    def set_device_ib_mac(self, pci_slot, mac=None):
        """Set IB mac address for given virtual function address

        :param pci_slot: VF pci address
        :param mac: MAC address for the VF, if None then the device
                    mac will be invalidated.

        Note:
        When setting a valid MAC address, the VF admin state will change to
        enable. When invalidating MAC address (mac=None),
        VF admin state will change to disable.
        """
        embedded_switch = self._get_emb_eswitch(pci_slot)
        if embedded_switch:
            embedded_switch.set_device_ib_mac(pci_slot, mac)

    def config_vlan_ib(self, pci_slot, vlan):
        """Set VF IB VLAN (Set Infiniband PKEY for VF)

        :param pci_slot: VF PCI address
        :param vlan: VLAN ID
        """
        embedded_switch = self._get_emb_eswitch(pci_slot)
        if embedded_switch:
            embedded_switch.config_vlan_ib(pci_slot, vlan)

    def _process_emb_switch_map(self, phys_net, dev_name):
        """Process emb_switch_map
        :param phys_net: physical network
        :param dev_name: device name
        """
        emb_switches = self.emb_switches_map.get(phys_net, [])
        for switch in emb_switches:
            if switch.dev_name == dev_name:
                # Embedded switch already created skip
                return

        # Create a new Embedded switch

        if (self.pci_utils.is_sriov_pf(dev_name) and
                self.pci_utils.verify_vendor_pf(dev_name, constants.VENDOR)):
            self._create_emb_switch(phys_net, dev_name)
        else:
            LOG.error("PF %s must have Mellanox Vendor ID, "
                      "SR-IOV and driver module enabled. "
                      "Terminating!", dev_name)
            raise SystemExit(1)

    def discover_devices(self, fabrics):
        """Discover which Virtual functions to manage.

        Discover devices, and create embedded switch object for network device
        :param fabrics: list of <physnet, pf netdev> tuples
        """
        for phys_net, dev_name in fabrics:
            self._process_emb_switch_map(phys_net, dev_name)

    def _create_emb_switch(self, phys_net, dev_name):
        embedded_switch = EmbSwitch(dev_name)
        # TODO(adrianc): eswitch service should not need to set PF netdev
        # link state. This should be handled during deployment. In the furure,
        # consider removing config_port_up.
        embedded_switch.pf_dev_wrapper.config_port_up()

        self.emb_switches_map.setdefault(phys_net, []).append(embedded_switch)
        for pci_slot in embedded_switch.get_pci_slot_list():
            self.pci_slot_map[pci_slot] = embedded_switch

    def _get_emb_eswitch(self, pci_slot):
        """Get embedded switch.

        Get embedded switch by pci slot
        :param pci_slot: pci slot
        """
        embedded_switch = self.pci_slot_map.get(pci_slot)
        return embedded_switch



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex
import sys

from neutron.conf.agent import common as common_conf
from oslo_config import cfg
from oslo_log import log as logging
from oslo_privsep import priv_context

LOG = logging.getLogger(__name__)
DEFAULT_INTERFACE_MAPPINGS = []

mlx_daemon_opts = [
    cfg.StrOpt('socket_os_transport', default="tcp"),
    cfg.StrOpt('socket_os_port', default="60001"),
    cfg.StrOpt('socket_os_addr', default="0.0.0.0"),
    cfg.ListOpt('fabrics',
                default=DEFAULT_INTERFACE_MAPPINGS,
                help=("List of <physical_network>:<physical_interface>")),
    cfg.IntOpt('default_timeout',
               default=5000,
               help=('Default timeout waiting for messages')),
    cfg.IntOpt('max_polling_count',
               default=5,
               help=('Daemon will do sync after max_polling_count '
                     '* default_timeout')),
]


def register_mlnx_daemon_opts(conf=cfg.CONF):
    conf.register_opts(mlx_daemon_opts, "DAEMON")


def register_opts(conf=cfg.CONF):
    # NOTE(adrianc) Logging opts are registered during import of
    # common_conf module.
    common_conf.register_root_helper(conf)
    register_mlnx_daemon_opts(conf)


def setup_logging(conf=cfg.CONF):
    """Sets up the logging options for a log with supplied name."""
    logging.setup(conf, 'eswitchd')
    LOG.info("Logging enabled!")
    LOG.info("%(prog)s Started!",
             {'prog': sys.argv[0]})
    LOG.debug("command line: %s", " ".join(sys.argv))


def get_root_helper(conf=cfg.CONF):
    return conf.AGENT.root_helper


def setup_privsep(conf=cfg.CONF):
    priv_context.init(root_helper=shlex.split(get_root_helper(conf)))


def init(args, **kwargs):
    cfg.CONF(args=args, project='eswitchd',
             **kwargs)
    register_opts(cfg.CONF)
    setup_logging(cfg.CONF)
    setup_privsep(cfg.CONF)
